import os
import numpy as np
import yaml
from copy import deepcopy
import warnings
import matplotlib.pyplot as plt
import networkx as nx
from typing import List, Union, Dict, Optional, Any, Type, Tuple
import eagerx.core.log as log
from eagerx.utils.utils import (
    load,
    is_compatible,
    supported_types,
)
from eagerx.utils.utils_sub import substitute_args
from eagerx.utils.network_utils import (
    reset_graph,
    episode_graph,
    plot_graph,
    color_nodes,
    color_edges,
    is_stale,
)
import eagerx
from eagerx.core.space import Space
from eagerx.core.entities import Node
from eagerx.core.view import SpecView
from eagerx.core.specs import (
    ResetNodeSpec,
    ObjectSpec,
    EngineSpec,
    ProcessorSpec,
    EntitySpec,
    NodeSpec,
)

yaml.Dumper.ignore_aliases = lambda *args: True  # todo: check if needed.


def merge(a, b, path=None):
    """merges b into a"""
    # If it is a spec, convert to params
    if path is None:
        path = []
    for key in b:
        if isinstance(b[key], EntitySpec):
            b[key] = b[key].params
        if isinstance(b[key], SpecView):
            b[key] = b[key].to_dict()
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


class Graph:
    """The Graph API allows users to form a graph of connected nodes and objects."""

    def __init__(self, state: Dict):
        self._state = state

    def __str__(self):
        return yaml.dump(self._state)

    @classmethod
    def create(
        cls,
        nodes: Optional[Union[Union[NodeSpec, ResetNodeSpec], List[Union[NodeSpec, ResetNodeSpec]]]] = None,
        objects: Optional[Union[ObjectSpec, List[ObjectSpec]]] = None,
    ) -> "Graph":
        """Create a new graph with nodes and objects.

        :param nodes: Nodes to add.
        :param objects: Objects to add.
        :return: The graph.
        """
        if nodes is None:
            nodes = []
        if objects is None:
            objects = []
        if isinstance(nodes, (NodeSpec, ResetNodeSpec)):
            nodes = [nodes]
        if isinstance(objects, ObjectSpec):
            objects = [objects]

        # Add action & observation node to list
        from eagerx.core.nodes import EnvNode

        environment = EnvNode.make()
        nodes += [environment]

        # Create a state
        state = dict(nodes=dict(), connects=list(), backup=dict(), gui_state=dict(), engine_id=None)
        graph = cls(state)
        graph.add(nodes)
        graph.add(objects)
        return graph

    def add(
        self,
        entities: Union[
            Union[NodeSpec, ResetNodeSpec, ObjectSpec, EngineSpec],
            List[Union[NodeSpec, ResetNodeSpec, ObjectSpec, EngineSpec]],
        ],
    ) -> None:
        """Add nodes/objects to the graph.

        :param entities: Nodes/objects to add.
        """
        Graph._add(self._state, entities)

    @staticmethod
    def _add(
        state: Dict,
        entities: Union[
            Union[NodeSpec, ResetNodeSpec, ObjectSpec, EngineSpec],
            List[Union[NodeSpec, ResetNodeSpec, ObjectSpec, EngineSpec]],
        ],
    ) -> None:
        """Add nodes/objects to the graph.

        :param entities: Nodes/objects to add.
        """
        if not isinstance(entities, list):
            entities = [entities]

        for entity in entities:
            # Check spec validity
            Graph._check_spec(entity)

            name = entity.config.name
            assert (
                name not in state["nodes"]
            ), f'There is already a node or object registered in this graph with name "{name}".'

            # Add node to state
            state["nodes"][name] = entity._params
            state["backup"][name] = entity.params

    def remove(self, names: Union[str, EntitySpec, List[Union[str, EntitySpec]]], remove: bool = False) -> List:
        """Removes nodes/objects from the graph.

        - First, all associated connections are disconnected.

        - Then, removes the node/object.

        :param names: Either the name or spec of the node/object that is to be removed.
        :param remove: Flag to also remove observations/actions if they are left disconnected after the node/object was removed.
                       Actions are only removed if they are completely disconnected.
        :return: list of disconnected connections.
        """
        if not isinstance(names, list):
            names = [names]
        disconnected = []
        for name in names:
            if isinstance(name, EntitySpec):
                name = name.params["config"]["name"]
                assert name in self._state["nodes"], f" No entity with name '{name}' in graph."
            for source, target in deepcopy(self._state["connects"]):
                if name in [source[0], target[0]]:
                    if source[0] == "environment":
                        action = source[2]
                        source = None
                    else:
                        action = None
                        source = self.get_view(source[0], source[1:])
                    if target[0] == "environment":
                        observation = target[2]
                        target = None
                    else:
                        observation = None
                        target = self.get_view(target[0], target[1:])
                    self.disconnect(source, target, action, observation, remove=remove)
                    disconnected.append([source, target])
            self._state["nodes"].pop(name)
            return disconnected

    def get_spec(self, name: str) -> EntitySpec:
        """Get Spec from the graph

        :param name: Name
        :return: The specification of the entity.
        """
        return Graph._get_spec(self._state, name)

    @staticmethod
    def _get_spec(state, name: str) -> EntitySpec:
        assert name in state["nodes"], f" No entity with name '{name}' in graph."
        params = state["nodes"][name]
        if "node_type" in params:  # Either Node or ResetNode
            if "targets" in params:  # == ResetNode
                spec = ResetNodeSpec(params)
            else:  # == Node
                spec = NodeSpec(params)
        else:  # == Object
            spec = ObjectSpec(params)
        return spec

    def add_component(
        self,
        entry: Optional[SpecView] = None,
        action: Optional[str] = None,
        observation: Optional[str] = None,
    ) -> None:
        """Selects an available component entry (e.g. input, output, etc...) that was not already selected.

        :param entry: Selects the entry, so that it can be connected.
        :param action: Adds a disconnected action entry.
        :param observation: Adds a disconnected observation entry.
        """
        # assert only action, only observation, only entry
        self._correct_signature(entry, action, observation)
        if entry is not None:  # component parameter
            self._add_component(entry)
        if action:
            self._add_action(action)
        if observation:
            self._add_observation(observation)

    def _add_component(self, entry: SpecView):
        """Adds a component entry to the selection list."""
        name, component, cname = entry()

        # For feedthroughs, add the corresponding output instead
        component = "outputs" if component == "feedthroughs" else component

        # Add cname to selection list if it is not already selected
        params = self._state["nodes"][name]
        assert cname not in params["config"][component], '"%s" already selected in "%s" under %s.' % (
            cname,
            name,
            component,
        )
        params["config"][component].append(cname)

    def _add_action(self, action: str):
        """Adds disconnected action entry to 'env/actions' node in self._state."""
        assert action != "set", 'Cannot define an action with the reserved name "set".'
        params_action = self._state["nodes"]["environment"]
        if action not in params_action["outputs"]:  # Action already registered
            params_action["outputs"][action] = dict()
            view = self.get_view("environment", ["outputs", action])
            self._add_component(view)

    def _add_observation(self, observation: str):
        """Adds disconnected observation entry to 'env/observations' node in self._state."""
        assert observation != "actions_set", 'Cannot define an observations with the reserved name "actions_set".'
        params_obs = self._state["nodes"]["environment"]
        if observation in params_obs["inputs"]:
            assert len(params_obs["inputs"][observation]) == 0, (
                'Observation "%s" already exists' " and is connected." % observation
            )
        else:
            params_obs["inputs"][observation] = dict()
            view = self.get_view("environment", ["inputs", observation])
            self._add_component(view)

    def remove_component(
        self,
        entry: Optional[SpecView] = None,
        action: Optional[str] = None,
        observation: Optional[str] = None,
        remove: bool = False,
    ) -> None:
        """Deselects a component entry (e.g. input, output, etc...) that was selected.

        - First, all associated connections are disconnected.

        - Then, deselects the component entry. For feedthroughs, it will also remove the corresponding output entry.

        :param entry: Deselects the entry.
        :param action: Removes an action entry.
        :param observation: Removes an observation entry
        :param remove: Flag to also remove observations/actions if they are left disconnected after the entry was removed.
                       Actions are only removed if they are completely disconnected.
        """
        self._correct_signature(entry, action, observation)
        if entry:  # component parameter
            self._remove_component(entry, remove=remove)
        if action:
            self._remove_action(action)
        if observation:
            self._remove_observation(observation)

    def _remove_component(self, entry: SpecView, remove: bool = False):
        """Removes a component entry from the selection list. It will first disconnect all connections in connect.
        For feedthroughs, it will remove the corresponding output from the selection list.
        """
        # For feedthroughs, remove the corresponding output instead
        name, component, cname = entry()
        component = "outputs" if component == "feedthroughs" else component
        entry = self.get_view(name, [component, cname])
        self._is_selected(self._state, entry)

        # Disconnect component entry
        self._disconnect_component(entry, remove=remove)

        # Remove cname from selection list
        params = self._state["nodes"][name]
        params["config"][component].remove(cname)

    def _remove_action(self, action: str):
        """Method to remove an action. It will first disconnect all connections in connect."""
        view = self.get_view("environment", ["outputs", action])
        self._remove_component(view, remove=False)
        params_action = self._state["nodes"]["environment"]
        source = ["environment", "outputs", action]
        connect_exists = False
        for c in self._state["connects"]:
            if source == c[0]:
                connect_exists = True
                target = c[1]
                break
        assert (
            not connect_exists
        ), f'Action entry "{action}" cannot be removed, because it is not disconnected. Connection with target {target} still exists.'
        params_action["outputs"].pop(action)

    def _remove_observation(self, observation: str):
        """Method to remove an observation. It will first disconnect all connections in connect."""
        view = self.get_view("environment", ["inputs", observation])
        self._remove_component(view, remove=False)
        params_obs = self._state["nodes"]["environment"]
        target = ["environment", "inputs", observation]
        connect_exists = False
        for c in self._state["connects"]:
            if target == c[1]:
                connect_exists = True
                source = c[0]
                break
        assert not connect_exists, (
            'Observation entry "%s" cannot be removed, because it is not disconnected. Connection with source %s still exists.'
            % (observation, source)
        )
        # assert observation in params_obs['inputs'], 'Observation "%s" cannot be removed,
        # because it does not exist.' % observation
        params_obs["inputs"].pop(observation)

    def connect(
        self,
        source: SpecView = None,
        target: SpecView = None,
        action: str = None,
        observation: str = None,
        window: Optional[int] = None,
        delay: Optional[float] = None,
        skip: Optional[bool] = None,
    ) -> None:
        """Connect an action/source (i.e. node/object component) to an observation/target (i.e. node/object component).

        :param source: Compatible source types are
                       :attr:`~eagerx.core.specs.NodeSpec.outputs`,
                       :attr:`~eagerx.core.specs.ObjectSpec.sensors`, and
                       :attr:`~eagerx.core.specs.ObjectSpec.states`.
        :param target: Compatible target types are
                       :attr:`~eagerx.core.specs.NodeSpec.inputs`,
                       :attr:`~eagerx.core.specs.ObjectSpec.actuators`,
                       :attr:`~eagerx.core.specs.ResetNodeSpec.targets`, and
                       :attr:`~eagerx.core.specs.ResetNodeSpec.feedthroughs`.
        :param action: Name of the action to connect (and add).
        :param observation: Name of the observation to connect (and add).
        :param window: A non-negative number that specifies the number of messages to pass to the node's :func:`~eagerx.core.entities.Node.callback`.

                       - *window* = 1: Only the last received input message.

                       - *window* = *x* > 1: The trailing last *x* received input messages.

                       - *window* = 0: All input messages received since the last call to the node's :func:`~eagerx.core.entities.Node.callback`.

                       .. note:: With *window* = 0, the number of input messages may vary and can even be zero.

        :param delay: A non-negative simulated delay (seconds). This delay is ignored if
                      :attr:`~eagerx.core.entities.Engine.simulate_delays` = True
                      in the engine's :func:`~eagerx.core.entities.Engine.spec`.
        :param skip: Skip the dependency on this input during the first call to the node's :func:`~eagerx.core.entities.Node.callback`.
                     May be necessary to ensure that the connected graph is directed and acyclic.
        """
        assert not source or not action, (
            'You cannot specify a source if you wish to connect action "%s",' " as the action will act as the source." % action
        )
        assert not target or not observation, (
            'You cannot specify a target if you wish to connect observation "%s",'
            " as the observation will act as the target." % observation
        )
        assert not (observation and action), "You cannot connect an action directly to an observation."

        if action:  # source = action
            self.add_component(action=action)
            self._connect_action(action, target)
            source = self.get_view("environment", ["outputs", action])
        elif observation:  # target = observation
            self.add_component(observation=observation)
            self._connect_observation(source, observation)
            target = self.get_view("environment", ["inputs", observation])
        self._connect(source, target, window, delay, skip)

    def _connect(
        self,
        source: SpecView = None,
        target: SpecView = None,
        window: Optional[int] = None,
        delay: Optional[float] = None,
        skip: Optional[bool] = None,
    ):
        """Method to connect a source to a target. For actions/observations, first a (new) disconnected entry must be created,
        after which an additional call to connect_action/observation is required before calling this method.
        For more info, see self.connect."""
        # Perform checks on source
        self._is_selected(self._state, source)

        # Perform checks on target
        target_name, target_comp, target_cname = target()
        if target_comp == "feedthroughs":
            assert window is None or window > 0, "Feedthroughs must have a window > 0, " "else no action can be fed through."
            assert not skip, (
                "Feedthroughs cannot skip, as they only feedthrough outputs. "
                "When setting skip=True, no msg can be fed through."
            )
            ft_view = self.get_view(target_name, ["outputs", target_cname])
            self._is_selected(self._state, ft_view)
        else:
            self._is_selected(self._state, target)

        # Make sure that source and target are compatible
        self._is_compatible(self._state, source, target)

        # Make sure that target is not already connected.
        for s, t in self._state["connects"]:
            t_name, t_comp, t_cname = t
            flag = not (target_name == t_name and target_comp == t_comp and target_cname == t_cname)
            assert flag, f'Target "{target}" is already connected to source "{s}"'

        # Add properties to target params
        if window is not None:
            self.set({"window": window}, target)
        if delay is not None:
            self.set({"delay": delay}, target)
        if skip is not None:
            self.set({"skip": skip}, target)

        # Add connection
        connect = [list(source()), list(target())]
        self._state["connects"].append(connect)

    def _connect_action(self, action, target):
        """Method to connect a (previously added) action, that *precedes* self._connect(source, target)."""
        params_action = self._state["nodes"]["environment"]
        assert action in params_action["outputs"], f'Action "{action}" must be added, before you can connect it.'
        name, component, cname = target()
        if component == "feedthroughs":
            component = "outputs"
            target = self.get_view(name, [component, cname])
            # Perform checks on target
            self._is_selected(self._state, target)
        else:
            # Perform checks on target
            self._is_selected(self._state, target)

        # Set properties in node params of 'env/actions'
        if len(params_action["outputs"][action]) > 0:  # Action already registered
            space_action = params_action["outputs"][action]["space"]
            space_target = target.space.to_dict()
            if not space_target == space_action:
                log.logwarn(
                    f'Conflicting space ({space_action}) for action "{action}". '
                    f"Not using the space ({space_target}) of {name}.{component}.{cname}"
                )
        else:
            assert target.space is not None, (
                f'"{name}.{component}.{cname}" does not have a space registered in the spec of "{name}". '
                f"Specify a space before connecting the action."
            )
            mapping = dict(
                rate="$(config rate)",
                processor=None,
                space=target.space,
            )
            self._set(params_action["outputs"], {action: mapping})

    def _connect_observation(self, source, observation):
        """Method to connect a (previously added & disconnected) observation, that *precedes* self._connect(source, target)."""

        params_obs = self._state["nodes"]["environment"]
        assert observation in params_obs["inputs"], f'Observation "{observation}" must be added, before you can connect it.'
        name, component, cname = source()

        assert source.space is not None, (
            f'"{name}.{component}.{cname}" does not have a space registered in the spec of "{name}". '
            f"Specify a space before connecting the observation."
        )

        # Set properties in node params of 'env/observations'
        assert len(params_obs["inputs"][observation]) == 0, 'Observation "%s" already connected.' % observation
        mapping = dict(
            delay=0.0,
            window=1,
            skip=False,
            processor=None,
            space=source.space,
            address=None,
        )
        self._set(params_obs["inputs"], {observation: mapping})

    def disconnect(
        self,
        source: SpecView = None,
        target: SpecView = None,
        action: str = None,
        observation: str = None,
        remove: bool = False,
    ) -> None:
        """Disconnects a source/action from a target/observation.

        :param source: Compatible source types are
                       :attr:`~eagerx.core.specs.NodeSpec.outputs`,
                       :attr:`~eagerx.core.specs.ObjectSpec.sensors`, and
                       :attr:`~eagerx.core.specs.ObjectSpec.states`.
        :param target: Compatible target types are
                       :attr:`~eagerx.core.specs.NodeSpec.inputs`,
                       :attr:`~eagerx.core.specs.ObjectSpec.actuators`,
                       :attr:`~eagerx.core.specs.ResetNodeSpec.targets`, and
                       :attr:`~eagerx.core.specs.ResetNodeSpec.feedthroughs`.
        :param action: Name of the action to connect (and add).
        :param observation: Name of the observation to connect (and add).
        :param remove: Flag to also remove observations/actions if they are left disconnected after the entry was disconnected.
                       Actions are only removed if they are completely disconnected.
        """
        self._disconnect(source, target, action, observation)
        if remove:
            if action:
                connect_exists = False
                source = ["environment", "outputs", action]
                for c in self._state["connects"]:
                    if source == c[0]:
                        connect_exists = True
                        break
                if not connect_exists:
                    self.remove_component(action=action)
            if observation:
                self.remove_component(observation=observation)

    def _disconnect(
        self,
        source: SpecView = None,
        target: SpecView = None,
        action: str = None,
        observation: str = None,
    ):
        """Disconnects a source from a target. The target is reset in self._state to its disconnected state."""
        assert not source or not action, (
            'You cannot specify a source if you wish to disconnect action "%s", '
            "as the action will act as the source." % action
        )
        assert not target or not observation, (
            'You cannot specify a target if you wish to disconnect observation "%s", '
            "as the observation will act as the target." % observation
        )
        assert not (observation and action), (
            "You cannot disconnect an action from an observation," " as such a connection cannot exist."
        )

        # Create source & target entries
        if action:
            assert target is not None, (
                f"If you want to disconnect action {action}, " f"please also specify the corresponding target."
            )

            source = self.get_view("environment", ["outputs", action])
        if observation:
            assert source is not None, (
                f"If you want to disconnect observation {observation}," f" please also specify the corresponding source."
            )

            target = self.get_view("environment", ["inputs", observation])

        # Check if connection exists
        self._is_selected(self._state, source)
        self._is_selected(self._state, target)

        # Check if connection exists
        connect_exists = False
        idx_connect = None
        for idx, c in enumerate(self._state["connects"]):
            if list(source()) == c[0] and list(target()) == c[1]:
                connect_exists = True
                idx_connect = idx
                break
        assert connect_exists, (
            f"The connection with source={source()} and target={target()} cannot be removed," " because it does not exist."
        )

        # Pop the connection from the state
        self._state["connects"].pop(idx_connect)

        # Reset source params to disconnected state
        if action:
            self._disconnect_action(action)
        else:
            pass  # Nothing to do here (for now)

        # Reset target params to disconnected state (reset to go back to default yaml), i.e. reset window/delay/skip.
        if observation:
            self._disconnect_observation(observation)
        else:
            target_name, target_comp, target_cname = target()
            target_params = self._state["nodes"][target_name]
            target_params[target_comp][target_cname] = self._state["backup"][target_name][target_comp][target_cname]

    def _disconnect_component(self, entry: SpecView, remove=False):
        """Disconnects all associated connects from self._state.
        **DOES NOT** remove observation entries if they are disconnected.
        **DOES NOT** remove action entries if they are disconnect and the last connection.
        """
        was_connected = False
        name, component, cname = entry()
        for source, target in deepcopy(self._state["connects"]):
            source = self.get_view(source[0], source[1:])
            target = self.get_view(target[0], target[1:])
            self._is_selected(self._state, source)
            self._is_selected(self._state, target)
            source_name, source_comp, source_cname = source()
            target_name, target_comp, target_cname = target()
            if source_name == "environment":
                action = source_cname
                source = None
            else:
                action = None
                source = source
            if target_name == "environment":
                observation = target_cname
                target = None
            else:
                observation = None
                target = target
            if name == source_name and component == source_comp and cname == source_cname:
                self.disconnect(source, target, action, observation, remove=remove)
                was_connected = True
            elif name == target_name and component == target_comp and cname == target_cname:
                self.disconnect(source, target, action, observation, remove=remove)
                was_connected = True
        return was_connected

    def _disconnect_action(self, action: str):
        """Returns the action entry back to its disconnected state.
        That is, remove space if it is not connected to any other targets."""
        params_action = self._state["nodes"]["environment"]
        assert action in params_action["outputs"], 'Cannot disconnect action "%s", as it does not exist.' % action
        source = ["environment", "outputs", action]
        connect_exists = False
        for c in self._state["connects"]:
            if source == c[0]:
                connect_exists = True
                break
        if not connect_exists:
            params_action["outputs"][action] = dict()

    def _disconnect_observation(self, observation: str):
        """Returns the observation entry back to its disconnected state (i.e. empty dict)."""
        params_obs = self._state["nodes"]["environment"]
        assert observation in params_obs["inputs"], 'Cannot disconnect observation "%s", as it does not exist.' % observation
        params_obs["inputs"][observation] = dict()

    def rename(
        self,
        new: str,
        action: Optional[str] = None,
        observation: Optional[str] = None,
    ) -> None:
        """Renames an action/observation.

        :param new: New name.
        :param action: Old action name.
        :param observation: Old observation name.
        """
        if action:
            assert observation is None, "Cannot supply both an action and observation."
            entry = self.get_view("environment", ["outputs", action])
        if observation:
            assert action is None, "Cannot supply both an action and observation."
            assert observation is not None, "Either an action or observation must be supplied."
            entry = self.get_view("environment", ["inputs", observation])
        self._rename_component(entry, new_cname=new)

    def _rename_component(self, entry: SpecView, new_cname: str):
        """Renames the component name (cname) of an entity (node/object) in _state['nodes'] and self._state[connects].
        We cannot change names for node/object components, because their python implementation could depend on it.
        Does not work for feedthroughs.
        """
        name, component, old_cname = entry()
        default = self._state["backup"][name]
        params = self._state["nodes"][name]

        # For now, we only support changing action/observation cnames
        assert name in ["environment", "environment"], (
            f'Cannot change "{old_cname}" of "{name}". Only name ' "changes to observations and actions are supported."
        )
        assert new_cname not in params[component], f'"{new_cname}" already defined in "{name}" under {component}.'

        # Rename cname in params
        # Does not work for outputs with feedthroughs. Then, both outputs and feedthroughs cnames must be changed.
        for d in (params, default):
            if component in d and old_cname in d[component]:
                assert new_cname not in d[component], f'"{new_cname}" already defined in "{name}" under {component}.'
                d[component][new_cname] = d[component].pop(old_cname)
            if component in d["config"] and old_cname in d["config"][component]:
                assert new_cname not in d["config"][component], f'"{new_cname}" already defined in "{name}" under {component}.'
                d["config"][component].remove(old_cname)
                d["config"][component].append(new_cname)

        # Rename cname in all connects
        for source, target in self._state["connects"]:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target

            if source_comp == component and source_cname == old_cname:
                source[2] = new_cname
            if target_comp == component and target_cname == old_cname:
                target[2] = new_cname

    def set(
        self,
        mapping: Any,
        entry: Optional[SpecView] = None,
        action: Optional[str] = None,
        observation: Optional[str] = None,
        parameter: Optional[str] = None,
    ) -> None:
        """Sets the parameters of a node/object/action/observation.

        :param mapping: Either a mapping with *key* = *parameter*,
                        or a single value that corresponds to the optional *parameter* argument.
        :param entry: The entry whose parameters are mutated.
        :param action: Action name whose parameters are mutated.
        :param observation: observation name whose parameters are mutated.
        :param parameter: If only a single value needs to be set. See documentation for *mapping*.
        """
        self._correct_signature(entry, action, observation)
        if action:
            entry = self.get_view("environment", ["outputs", action])
        if observation:
            entry = self.get_view("environment", ["inputs", observation])
        if parameter is None:
            if isinstance(mapping, SpecView):
                mapping = mapping.to_dict()
            assert isinstance(mapping, dict), "Can only set mappings of type dict. Else, also set 'parameter=<param_name>'."
        else:
            mapping = {parameter: mapping}

        for parameter, value in mapping.items():
            if parameter:
                getattr(entry, parameter)  # Check if parameter exists
            if parameter == "processor":
                self._set_processor(entry, value)
            else:
                t = entry()
                name = t[0]
                if t[1] == "config":
                    assert parameter not in [
                        "sensors",
                        "actuators",
                        "targets",
                        "states",
                        "inputs",
                        "outputs",
                    ], "You cannot modify component parameters with this function. Use _add/remove_component(..) instead."
                    assert parameter not in ["name"], f"You cannot rename '{name}'."
                    p = self._state["nodes"][name]["config"]
                else:
                    name, component, cname = entry()
                    p = self._state["nodes"][name][component][cname]
                self._set(p, {parameter: value})

    def _set_processor(self, entry: SpecView, processor: Dict):
        """Replaces the processor specified for a node's/object's input."""
        if isinstance(processor, ProcessorSpec):
            processor = processor.params
        elif isinstance(processor, SpecView):
            processor = processor.to_dict()

        _ = entry.processor  # Check if parameter exists

        # Replace processor
        name, component, cname = entry()
        params = self._state["nodes"][name]
        params[component][cname]["processor"] = processor

    def get(
        self,
        entry: Optional[Union[SpecView, EntitySpec]] = None,
        action: Optional[str] = None,
        observation: Optional[str] = None,
        parameter: Optional[str] = None,
    ) -> Any:
        """Fetches the parameters of a node/object/action/observation.

        :param entry: The entry whose parameters are fetched.
        :param action: Action name whose parameters are fetched.
        :param observation: observation name whose parameters are fetched.
        :param parameter: If only a single parameter needs to be fetched.
        :return: Parameters
        """
        if isinstance(entry, EntitySpec):
            name = entry.params["config"]["name"]
            assert name in self._state["nodes"], f" No entity with name '{name}' in graph."
            return self._state["nodes"][name]
        self._correct_signature(entry, action, observation)
        if action:
            entry = self.get_view("environment", ["outputs", action])
        if observation:
            entry = self.get_view("environment", ["inputs", observation])
        if parameter:
            return getattr(entry, parameter)
        else:
            return entry

    def register(
        self, rate: float, engine: Optional[EngineSpec] = None
    ) -> Tuple["Graph", NodeSpec, EngineSpec, List[NodeSpec], Optional[NodeSpec]]:
        state = deepcopy(self._state)
        engine = deepcopy(engine)
        return self._register(state, rate, engine)

    @staticmethod
    def _register(
        state, rate: float, engine: Optional[EngineSpec] = None
    ) -> Tuple["Graph", NodeSpec, EngineSpec, List[NodeSpec], Optional[NodeSpec]]:
        assert rate >= 0, f"The rate must be non-negative ('{rate}' !=> 0)."

        # Add engine
        if engine:
            Graph._add_engine(state, engine)

        # Check that there is an engine specified.
        assert state["engine_id"] is not None, "Cannot register graph. There was no engine specified."

        # Substitute all objects
        Graph._substitute_all_objects(state)

        # Check if valid graph.
        assert Graph._is_valid(state, plot=False), "Graph not valid."

        # Reload all entities
        Graph._reload(state)

        # Add addresses based on connections
        for source, target in state["connects"]:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target

            # For actions & observations, replace default args
            if source_name == "environment":
                default = state["nodes"][target_name]["config"]
                context = {"config": default}
                cname_params = state["nodes"][source_name][source_comp][source_cname]
                substitute_args(cname_params, context, only=["config", "ns"])
                cname_params["rate"] = "$(config rate)"  # Do not copy the rate.
                address = "%s/%s/%s" % ("environment", source_comp, source_cname)
            elif target_name == "environment":
                default = state["nodes"][source_name]["config"]
                context = {"config": default}
                cname_params = state["nodes"][target_name][target_comp][target_cname]
                substitute_args(cname_params, context, only=["config", "ns"])
                address = "%s/%s/%s" % (source_name, source_comp, source_cname)
            else:
                address = "%s/%s/%s" % (source_name, source_comp, source_cname)
            state["nodes"][target_name][target_comp][target_cname]["address"] = address

            # check if dtypes match
            s = Graph._get_view(state, source[0], source[1:])
            t = Graph._get_view(state, target[0], target[1:])
            Graph._is_compatible(state, s, t)
            source_dtype = state["nodes"][source_name][source_comp][source_cname]["space"]["dtype"]
            state["nodes"][target_name][target_comp][target_cname]["dtype"] = source_dtype

        # Extract node_names
        node_names = ["env/supervisor"]
        target_addresses = []
        for _i, params in state["nodes"].items():
            if params["config"]["name"] == "engine":
                continue
            node_names.append(params["config"]["name"])
            if "targets" in params["config"]:
                for cname in params["config"]["targets"]:
                    address = params["targets"][cname]["address"]
                    target_addresses.append(address)
        assert "node_names" not in engine.config, f'Keyword "{"node_names"}" is a reserved keyword.'
        assert "target_addresses" not in engine.config, f'Keyword "{"target_addresses"}" is a reserved keyword.'
        msg = "An Engine can only be launched with the options {ENVIRONMENT, EXTERNAL, NEW_PROCESS}."
        assert not engine.config.process == eagerx.ENGINE, msg
        with engine.config as d:
            d.node_names = node_names
            d.target_addresses = target_addresses

        # Get specs
        environment, engine, nodes, render = Graph._get_all_node_specs(state)

        # Set environment rate
        environment.config.rate = rate
        return Graph(deepcopy(state)), environment, engine, nodes, render

    @staticmethod
    def _get_all_node_specs(state: Dict) -> Tuple[NodeSpec, EngineSpec, List[NodeSpec], Optional[NodeSpec]]:
        # Get specs
        nodes = []
        render: Optional[NodeSpec] = None
        engine: EngineSpec = None
        environment: NodeSpec = None
        for name, params in state["nodes"].items():
            assert "node_type" in params, f"Node `{name}` is not a valid node. Have all Objects been replaced by nodes?"
            if name == "environment":
                environment = NodeSpec(params)
            elif name == "env/render":
                render = NodeSpec(params)
                nodes.append(render)
            elif name == "engine":
                engine = EngineSpec(params)
            else:
                nodes.append(NodeSpec(params))
        assert environment, "No environment node defined in the graph."
        return environment, engine, nodes, render

    def render(
        self,
        source: SpecView,
        rate: float,
        processor: Optional[ProcessorSpec] = None,
        window: Optional[int] = None,
        delay: Optional[float] = None,
        skip: Optional[bool] = None,
        render_cls: Type[Node] = None,
        process: int = eagerx.process.NEW_PROCESS,
        encoding: str = "bgr",
        **kwargs,
    ):
        """Visualize rgb images produced by a node/sensor in the graph. The rgb images must be of `dtype=uint8` and
        `shape=(height, width, 3)`.

        :param source: Compatible source types are :attr:`~eagerx.core.specs.NodeSpec.outputs` and
                       :attr:`~eagerx.core.specs.ObjectSpec.sensors`.
        :param rate: The rate (Hz) at which to render the images.
        :param processor: Processes the received message before passing it
                          to the target node's :func:`~eagerx.core.entities.Node.callback`.
        :param window: A non-negative number that specifies the number of messages to pass to the node's :func:`~eagerx.core.entities.Node.callback`.

                       - *window* = 1: Only the last received input message.

                       - *window* = *x* > 1: The trailing last *x* received input messages.

                       - *window* = 0: All input messages received since the last call to the node's :func:`~eagerx.core.entities.Node.callback`.

                       .. note:: With *window* = 0, the number of input messages may vary and can even be zero.

        :param delay: A non-negative simulated delay (seconds). This delay is ignored if
                      :attr:`~eagerx.core.entities.Engine.simulate_delays` = True
                      in the engine's :func:`~eagerx.core.entities.Engine.spec`.
        :param skip: Skip the dependency on this input during the first call to the node's :func:`~eagerx.core.entities.Node.callback`.
                     May be necessary to ensure that the connected graph is directed and acyclic.
        :param render_cls: The :attr:`~eagerx.core.entities.Node` of the render node.
                           By default, it uses the standard `RenderNode`. In Google colab, the `ColabRender` class is used.
        :param process: Process in which the render node is launched. See :class:`~eagerx.core.constants.process` for all
                        options.
        :param encoding: The encoding (`bgr` or `rgb`) of the render source.
        :param kwargs: Optional arguments required by the render node.
        """
        # Delete old render node from self._state['nodes'] if it exists
        if "env/render" in self._state["nodes"]:
            self.remove("env/render")

        # Add (new) render node to self._state['node']
        if render_cls is None:
            if bool(eval(os.environ.get("EAGERX_COLAB", "0"))):
                from eagerx.core.nodes import ColabRender as render_cls

                process = eagerx.process.ENVIRONMENT
            else:
                from eagerx.core.nodes import RenderNode as render_cls
        render = render_cls.make(rate=rate, process=process, encoding=encoding, **kwargs)
        # todo: How to change space of render.inputs.image when a processor is added.
        render.inputs.image.space = source.space
        self.add(render)

        # Create connection
        target = self.get_view("env/render", depth=["inputs", "image"])
        target.processor = processor
        self.connect(
            source=source,
            target=target,
            window=window,
            delay=delay,
            skip=skip,
        )

    def save(self, file: str) -> None:
        """Saves the graph state.

        The state is saved in *.yaml* format and contains the state of every added node, object, action, and observation
        and the connections between them.

        :param file: A string giving the name (and the file if the file isn't in the current working directory).
        """
        with open(file, "w") as outfile:
            yaml.dump(self._state, outfile, default_flow_style=False)

    @classmethod
    def load(cls, file: str):
        """Loads the graph state.

        The state is loaded in *.yaml* format and contains the state of every added node, object, action, and observation
        and the connections between them.

        :param file: A string giving the name (and the file if the file isn't in the current working directory).
        """
        must_raise = False
        with open(file, "r") as stream:
            try:
                state = yaml.safe_load(stream)
                if "engine_id" not in state:
                    warnings.warn("Old graph fromat. Adding `engine` key to state.")
                    state["engine_id"] = None
            except yaml.YAMLError as exc:
                must_raise = True
                print(exc)
        if must_raise:
            raise yaml.YAMLError(f"Could not load graph from `{file}`.")
        else:
            # Import all Objects & Nodes (to register & see availability).
            for key, params in state["nodes"].items():
                try:
                    load(params["config"]["entity_id"])
                except AttributeError as e:
                    warnings.warn(f"Could not load `{key}`: {e}")
            return cls(state)

    def reload(self):
        """Reloads (ie imports) all entities in the graph."""
        Graph._reload(self._state)

    @staticmethod
    def _reload(state: Dict):
        """Reloads (ie imports) all entities in the graph."""
        for key, params in state["nodes"].items():
            try:
                load(params["config"]["entity_id"])
            except AttributeError as e:
                raise ImportError(f"Could not load `{key}`: {e}")

    def gui(
        self, interactive: Optional[bool] = True, resolution: Optional[List[int]] = None, filename: Optional[str] = None
    ) -> Union[None, np.ndarray]:
        """Opens a graphical user interface of the graph.

        .. note:: Requires `eagerx-gui`:

                .. highlight:: python
                .. code-block:: python

                    pip3 install eagerx-gui

        :param interactive: If `True`, an interactive application is launched.
                            Otherwise, an RGB render of the GUI is returned.
                            This could be useful when using a headless machine.
        :param resolution: Specifies the resolution of the returned render when `interactive` is `False`.
                           If `interactive` is `True`, this argument is ignored.
        :param filename: If provided, the GUI is rendered to an svg file with this name.
                         If `interactive` is `True`, this argument is ignored.
        :return: RGB render of the GUI if `interactive` is `False`.
        """
        # Make a copy of the state
        state = deepcopy(self._state)

        # Replace environment node with environment node
        Graph._substitute_environment_node(state)

        try:
            from eagerx_gui import launch_gui, render_gui
        except ImportError as e:
            log.logwarn(
                f"{e}. You will likely have to install eagerx-gui. It can be installed by running: pip install eagerx-gui"
            )
            return

        # Launch gui
        if interactive:
            # Only overwrite gui_state (node locations, etc...)
            # Note: `gui_state` may have info on nodes 'env/actions' and 'env/observations',
            #       while the graph state actually only includes node `environment`.
            self._state["gui_state"] = launch_gui(state)["gui_state"]
        else:
            return render_gui(state, resolution=resolution, filename=filename)

    @staticmethod
    def _is_selected(state: Dict, entry: SpecView):
        """Check if provided entry was selected in params."""
        name, component, cname = entry()
        assert name in state["nodes"], f'No entity with the anme "{name}" added to this graph.'
        params = state["nodes"][name]
        component = "outputs" if component == "feedthroughs" else component
        assert cname in params["config"][component], f'"{cname}" not selected in "{name}" under "config" in {component}.'

    @staticmethod
    def _is_compatible(state: Dict, source: SpecView, target: SpecView):
        """Check if provided the provided entries are compatible."""
        # Valid source and target components
        targets = ["inputs", "actuators", "feedthroughs", "targets"]
        sources = ["outputs", "states", "sensors"]

        # Provided entries
        target_name, target_component, target_cname = target()
        source_name, source_component, source_cname = source()
        base_msg = (
            f"'{target_name}.{target_component}.{target_cname}' cannot be connected with "
            f"'{source_name}.{source_component}.{source_cname}')."
        )

        # Check if target & source are validly chosen
        assert target_component in targets, f"{base_msg} '{target_component}' cannot be a target."
        assert source_component in sources, f"{base_msg} '{source_component}' cannot be a source."

        # Check if combinations are valid.
        if source_component in ["outputs", "sensors"]:
            valid = ["inputs", "feedthroughs", "actuators"]
            msg = f"{base_msg} '{source_component}' can only be connected to any of the components in '{valid}'."
            assert target_component in valid, msg
        else:  # source_component == "states":
            valid = ["targets"]
            msg = f"{base_msg} '{source_component}' can only be connected to any of the components in '{valid}'."
            assert target_component in valid, msg

            # Check that the state corresponds to an object (i.e. is an engine state)
            msg = (
                f"{base_msg} Only '{source_component}' of Objects can be connected to targets. '{source_name}' "
                "is not of type Object."
            )
            assert "node_type" not in state["nodes"][source_name], msg

        # Check if dtypes match
        source_params = state["nodes"][source_name][source_component][source_cname]
        target_params = state["nodes"][target_name][target_component][target_cname]
        assert (
            source_params["space"] is not None
        ), f"source={source_name}.{source_component}.{source_cname} does not have a space defined."
        assert (
            target_params["space"] is not None
        ), f"target={target_name}.{target_component}.{target_cname} does not have a space defined."
        source_dtype = source_params["space"]["dtype"]
        target_dtype = target_params["space"]["dtype"]
        try:
            is_compatible(source_dtype, target_dtype)
        except AssertionError as e:
            msg = (
                f"Incorrect connection of (source={source_name}.{source_component}.{source_cname}) with "
                f"(target={target_name}.{target_component}.{target_cname}): {e}"
            )
            raise IOError(msg)

    @staticmethod
    def _correct_signature(
        entry: Optional[SpecView] = None,
        action: Optional[str] = None,
        observation: Optional[str] = None,
    ):
        # assert only action, only observation, or only entry
        if entry:  # component parameter
            assert action is None, "If 'entry' is specified, action argument cannot be specified."
            assert observation is None, "If 'entry' is specified, observation argument cannot be specified."
        if action:
            assert observation is None, "If action is specified, observation must be None."
            assert entry is None, "If action is specified, the 'entry' argument cannot be specified."
        if observation:
            assert action is None, "If observation is specified, action must be None."
            assert entry is None, "If action is specified, the 'entry' argument cannot be specified."

    def is_valid(self, plot=True) -> bool:
        """Checks the validity of the graph.

        - Checks if all selected actions, observations,
          :attr:`~eagerx.core.specs.NodeSpec.inputs`,
          :attr:`~eagerx.core.specs.ObjectSpec.actuators`,
          :attr:`~eagerx.core.specs.ResetNodeSpec.targets`, and
          :attr:`~eagerx.core.specs.ResetNodeSpec.feedthroughs` are connected.

        - Checks if the graph is directed and acyclic.

        :param plot: Flag to plot the graph. Can be helpful to identify cycles in the graph that break the required acyclic property.
        :returns: flag that specifies the validity of the graph.
        """
        return self._is_valid(self._state, plot=plot)

    @staticmethod
    def _is_valid(state, plot=True):
        state = deepcopy(state)
        Graph.check_inputs_have_address(state)
        Graph.check_graph_is_acyclic(state, plot=plot)
        return True

    @staticmethod
    def check_inputs_have_address(state):
        state = deepcopy(state)
        for source, target in state["connects"]:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target
            address = "%s/%s/%s" % (source_name, source_comp, source_cname)
            state["nodes"][target_name][target_comp][target_cname]["address"] = address

        for name, params in state["nodes"].items():
            if "node_type" in params:
                for component in params["config"]:
                    if component not in [
                        "inputs",
                        "outputs",
                        "targets",
                        "feedthroughs",
                        "states",
                    ]:
                        continue
                    for cname in params["config"][component]:
                        assert cname in params[component], (
                            f'"{cname}" was selected in {component} of "{name}", ' "but has no implementation."
                        )
                        if component not in ["inputs", "targets", "feedthroughs"]:
                            continue
                        p = params[component][cname]
                        assert "address" in p and p["address"] is not None, (
                            f'"{cname}" was selected in {component} '
                            f'of "{name}", but no address could be '
                            "produced/inferred. This likely means that it has not been connected. "
                            "Either deselect it, or connect it."
                        )
            else:
                for component in params["config"]:
                    if component not in ["sensors", "actuators", "states"]:
                        continue
                    for cname in params["config"][component]:
                        assert cname in params[component], (
                            f'"{cname}" was selected in {component} of "{name}", ' "but has no (agnostic) implementation."
                        )
                        if component not in ["actuators"]:
                            continue
                        assert "address" in params[component][cname], (
                            f'"{cname}" was selected in {component} of "{name}", '
                            "but no address could be produced/inferred. This likely means that it has not been connected. "
                            "Either deselect it, or connect it."
                        )
        return True

    @staticmethod
    def check_graph_is_acyclic(state, plot=True):
        # Add nodes
        G = nx.MultiDiGraph()
        # label_mapping = {
        #     "env/render/done": "render",
        # }
        label_mapping = dict()
        for node, params in state["nodes"].items():
            default = params["config"]
            if "node_type" not in state["nodes"][node]:  # Object
                if "sensors" in default:
                    for cname in default["sensors"]:
                        name = "%s/sensors/%s" % (node, cname)
                        G.add_node(
                            name,
                            remain_active=False,
                            always_active=True,
                            is_stale=False,
                            has_tick=False,
                        )
                        if not ("actuators" in default and cname in default["actuators"]):
                            # label_mapping[name] = "%s/%s" % (node, cname)
                            label_mapping[name] = str(len(label_mapping))
                if "actuators" in default:
                    for cname in default["actuators"]:
                        name = "%s/actuators/%s" % (node, cname)
                        G.add_node(
                            name,
                            remain_active=True,
                            always_active=False,
                            is_stale=False,
                            has_tick=False,
                        )
                        if not ("sensors" in default and cname in default["sensors"]):
                            # label_mapping[name] = "%s/%s" % (node, cname)
                            label_mapping[name] = str(len(label_mapping))
            else:  # node
                for cname in default["outputs"]:
                    name = "%s/%s" % (node, cname)
                    label_mapping[name] = str(len(label_mapping))  # if node not in ["environment", "engine"] else name
                    G.add_node(
                        name,
                        remain_active=True if "tick" in default["inputs"] else False,
                        always_active=True if node == "engine" and cname == "tick" else False,
                        is_stale=False,
                        has_tick=False,
                    )

        # Add edges
        target_comps = ["inputs", "actuators", "feedthroughs"]
        source_comps = ["outputs", "sensors"]
        for source, target in state["connects"]:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target
            # Do not consider engine inputs for DAG check.
            if target_name == "engine":
                continue
            if source_comp in source_comps and target_comp in target_comps:
                # Determine source node name
                if "node_type" in state["nodes"][source_name]:
                    source_edge = "%s/%s" % (source_name, source_cname)
                else:
                    source_edge = "%s/%s/%s" % (source_name, source_comp, source_cname)
                # Determine target node name
                target_edges = []
                target_default = state["nodes"][target_name]["config"]
                if "node_type" in state["nodes"][target_name]:
                    for cname in target_default["outputs"]:
                        target_edge = "%s/%s" % (target_name, cname)
                        target_edges.append(target_edge)
                else:
                    target_edge = "%s/%s/%s" % (target_name, target_comp, target_cname)
                    target_edges.append(target_edge)

                # Determine stale nodes in real_reset routine via feedthrough edges
                if target_comp == "feedthroughs":
                    feedthrough = True
                else:
                    feedthrough = False

                # Determine edges that do not break DAG property (i.e. edges that are skipped)
                skip = state["nodes"][target_name][target_comp][target_cname]["skip"]
                color = "green" if skip else "black"
                style = "dotted" if skip else "solid"

                # Add edge
                for target_edge in target_edges:
                    G.add_edge(
                        source_edge,
                        target_edge,
                        color=color,
                        feedthrough=feedthrough,
                        style=style,
                        alpha=1.0,
                        is_stale=False,
                        skip=skip,
                        source=source,
                        target=target,
                    )

        # Color nodes based on in/out going edges
        not_active = is_stale(G)
        color_nodes(G)
        color_edges(G)

        # Remap action & observation labels to more readable form
        # G = nx.relabel_nodes(G, label_mapping)

        # Check if graph is acyclic (excluding 'skip' edges)
        H, cycles = episode_graph(G)
        is_dag = nx.is_directed_acyclic_graph(H)

        # Plot graphs
        if plot:
            fig_env, ax_env = plt.subplots(nrows=1, ncols=1)
            ax_env.set_title("Communication graph (episode)")
            # Remap action & observation labels to more readable form
            G_plot = nx.relabel_nodes(deepcopy(G), label_mapping)
            _, _, _, pos = plot_graph(G_plot, k=2, ax=ax_env)
            plt.show()

        # Assert if graph is a directed-acyclical graph (DAG)
        cycle_strs = ["Circular dependency detected: "]
        for idx, connect in enumerate(cycles):
            connect.append(connect[0])
            s = " Loop %s: " % idx
            n = "\n" + "".join([" "] * len(s)) + "...-->"
            s = "\n\n" + s + "...-->"
            for idx in range(len(connect) - 1):
                tmp, target = connect[idx]
                source, tmp2 = connect[idx + 1]
                source_name, source_comp, source_cname = source
                target_name, target_comp, target_cname = target
                assert source_name == target_name, "Source and target not equal: %s, %s" % (source, target)
                connect_name = "%s/%s/%s][%s/%s/%s" % tuple(list(source) + list(target))
                node_name = ("Node: " + source_name).center(len(connect_name), " ")
                s += "[%s]-->" % connect_name
                n += "[%s]-->" % node_name
            s += "..."
            n += "..."
            cycle_strs.append(s)
            cycle_strs.append(n)
            connect.pop(-1)
        assert is_dag, "".join(cycle_strs)
        assert len(not_active) == 0, (
            'Stale episode graph detected. Nodes "%s" will be stale, while they must be active (i.e. connected) in order for the graph to resolve (i.e. not deadlock).'
            % not_active
        )

        # Create a shallow copy graph that excludes feedthrough edges
        F = reset_graph(G)
        not_active = is_stale(F)
        color_nodes(F)
        color_edges(F)

        # Plot graphs
        if plot:
            fig_reset, ax_reset = plt.subplots(nrows=1, ncols=1)
            ax_reset.set_title("Communication graph (reset)")
            F_plot = nx.relabel_nodes(deepcopy(F), label_mapping)
            _, _, _, pos = plot_graph(F_plot, pos=pos, ax=ax_reset)
            plt.show()

        # Assert if reset graph is not stale
        has_real_reset = len([e for e, ft in nx.get_edge_attributes(G, "feedthrough").items() if ft]) > 0
        assert len(not_active) == 0 or not has_real_reset, (
            'Stale reset graph detected. Nodes "%s" will be stale, while they must be active (i.e. connected) in order for the graph to resolve (i.e. not deadlock).'
            % not_active
        )
        return True

    @staticmethod
    @supported_types(str, int, list, float, bool, dict, EntitySpec, SpecView, None)
    def _set(state, mapping):
        merge(state, mapping)

    @staticmethod
    def _get_view(state: Dict, name: str, depth: Optional[List[str]] = None):
        spec = Graph._get_spec(state, name)
        depth = depth if depth else []
        return SpecView(spec, depth=depth, name=name)

    def get_view(self, name: str, depth: Optional[List[str]] = None):
        return self._get_view(self._state, name, depth)

    @staticmethod
    def _check_spec(spec):
        if isinstance(spec, NodeSpec):
            from eagerx.core.entities import Node

            Node.check_spec(spec)
        elif isinstance(spec, ResetNodeSpec):
            from eagerx.core.entities import ResetNode

            ResetNode.check_spec(spec)
        elif isinstance(spec, ObjectSpec):
            from eagerx.core.entities import Object

            Object.check_spec(spec)

    def add_engine(self, engine: EngineSpec):
        Graph._add_engine(self._state, engine)

    @staticmethod
    def _add_engine(state, engine: EngineSpec):
        if state["engine_id"] is None:
            state["engine_id"] = engine.config.entity_id
            Graph._add(state, engine)
        else:
            raise ValueError("Cannot add engine. The graph already has an engine.")

    @staticmethod
    def _substitute_environment_node(state: Dict):
        """Substitute `environment` node with separate `env/actions` and `env/observations` node for improved visualization."""
        # Prepare environment spec
        environment = NodeSpec(state["nodes"]["environment"])

        # Add action & observation node.
        from eagerx.core.nodes import ActionsNode, ObservationsNode

        actions = ActionsNode.make()
        observations = ObservationsNode.make()
        Graph._add(state, [actions, observations])

        # Copy outputs to actions
        actions.config.outputs = environment.config.outputs
        with actions.outputs as a:
            a.update(environment.outputs.to_dict())

        # Copy inputs to observations
        observations.config.inputs = environment.config.inputs
        with observations.inputs as o:
            o.update(environment.inputs.to_dict())

        # Replace connections
        for s, t in state["connects"]:
            if s[:2] == ["environment", "outputs"]:
                s[0] = "env/actions"
            elif t[:2] == ["environment", "inputs"]:
                t[0] = "env/observations"

        # Remove environment node
        state["nodes"].pop("environment")
        state["backup"].pop("environment")

    @staticmethod
    def _substitute_all_objects(state: Dict):
        """state must already have an engine (ie _add_engine must be called first)."""
        assert state["engine_id"] is not None, "The graph state must have an engine, before objects can be substituted."
        for name in list(state["nodes"].keys()):
            # Skip if node
            if "node_type" in state["nodes"][name]:
                continue
            # Substitute object graph
            Graph._substitute_object(state, name)

    @staticmethod
    def _substitute_object(state: Dict, name: str):
        """state must already have an engine (ie _add_engine must be called first)."""
        assert state["engine_id"] is not None, "The graph state must have an engine, before objects can be substituted."

        # Get graph
        graph = Graph(state)

        # Get engine spec
        engine = EngineSpec(state["nodes"]["engine"])

        # Get object parameters
        params = state["nodes"][name]
        spec = ObjectSpec(params)

        # Make sure to register object by reimporting the object.
        # NOTE: This will only register all engine implementations that are defined within the Object's class definition.
        #       You may need to manually reload separate engine implementations when launching environments in subprocesses.
        load(spec.config.entity_id)

        # Register object's engine graph
        engine_graph = engine._register_object(spec)
        nodes, actuators, sensors, connects = engine_graph.register()

        # Construct context
        context = {"ns": {"obj_name": name}}
        substitute_args(params["config"], context, only=["ns"])  # First resolve args within the context
        substitute_args(params, context, only=["ns"])  # Resolve rest of params

        # Resolve namespaces
        substitute_args(nodes, context, only=["ns"])  # Resolve rest of params
        substitute_args(actuators, context, only=["ns"])  # Resolve rest of params
        substitute_args(sensors, context, only=["ns"])  # Resolve rest of params
        substitute_args(connects, context, only=["ns"])  # Resolve rest of params

        # Get agnostic & engine-specific definition
        agnostic = dict(actuators=params["actuators"], sensors=params["sensors"], states=params["states"])
        specific = dict(actuators=actuators, sensors=sensors)

        # Replace node names
        for key in list(nodes.keys()):
            key_sub = substitute_args(key, context, only=["ns"])
            nodes[key_sub] = nodes.pop(key)

        # Sensors & actuators
        dependencies = []
        for obj_comp in ["actuators", "sensors"]:
            for obj_cname in params["config"][obj_comp]:
                try:
                    entry_lst = specific[obj_comp][obj_cname]
                except KeyError:
                    raise KeyError(
                        f'"{obj_cname}" was selected in {obj_comp} of "{name}", but there is no implementation for it in engine "{state["engine_id"]}".'
                    )

                for entry in reversed(entry_lst):
                    node_name, node_comp, node_cname = entry["name"], entry["component"], entry["cname"]
                    obj_comp_params = agnostic[obj_comp][obj_cname]
                    node_params = nodes[node_name]

                    # Determine node dependency
                    dependencies += entry["dependency"]

                    # Set rate
                    rate = obj_comp_params["rate"]
                    msg_start = f'Different rate specified for {obj_comp[:-1]} "{obj_cname}" and enginenode "{node_name}": '
                    msg_end = "If an enginenode implements a sensor/actuator, their specified rates must be equal."
                    msg_mid = f'{node_params["config"]["rate"]} vs {rate}. '
                    assert node_params["config"]["rate"] == rate, msg_start + msg_mid + msg_end
                    for o in node_params["config"]["outputs"]:
                        msg_mid = f'{node_params["outputs"][o]["rate"]} vs {rate}. '
                        assert node_params["outputs"][o]["rate"] == rate, msg_start + msg_mid + msg_end

                    # Set component params
                    node_comp_params = nodes[node_name][node_comp][node_cname]
                    if obj_comp == "sensors":
                        # Make sure that space of node is contained within agnostic space.
                        agnostic_space = Space.from_dict(obj_comp_params["space"])
                        node_space = Space.from_dict(node_comp_params["space"])
                        msg = (
                            f"The space of EngineNode `{node_name}.{node_comp}.{node_cname}` is different "
                            f"(dtype, shape, low, or high) from the space of `{name}.{obj_comp}.{obj_cname}`: \n\n"
                            f"{node_name}.{node_comp}.{node_cname}.space={node_space} \n\n"
                            f"{name}.{obj_comp}.{obj_cname}.space={agnostic_space} \n\n"
                        )
                        assert agnostic_space.contains_space(node_space), msg
                        node_comp_params.update(obj_comp_params)
                        # node_comp_params["address"] = f"{name}/{obj_comp}/{obj_cname}" # todo: set to None?
                        # sensor_addresses[f"{node_name}/{node_comp}/{node_cname}"] = f"{name}/{obj_comp}/{obj_cname}"
                    else:  # Actuators
                        agnostic_space = Space.from_dict(obj_comp_params["space"])
                        node_space = Space.from_dict(node_comp_params["space"])
                        msg = (
                            f"The space of EngineNode `{node_name}.{node_comp}.{node_cname}` is different "
                            f"(dtype, shape, low, or high) from the space of `{name}.{obj_comp}.{obj_cname}`: \n\n"
                            f"{name}.{node_comp}.{node_cname}.space={node_space} \n\n"
                            f"{name}.{obj_comp}.{obj_cname}.space={agnostic_space} \n\n"
                        )
                        assert agnostic_space.contains_space(node_space), msg

                        agnostic_processor = obj_comp_params.pop("processor")
                        node_comp_params.update(obj_comp_params)

                        if agnostic_processor is not None:
                            msg = (
                                f"A processor was defined for {node_name}.{node_comp}.{node_cname}, however the engine "
                                "implementation also has a processor defined. You can only have one processor."
                            )
                            assert node_comp_params["processor"] is None, msg
                            node_comp_params["processor"] = agnostic_processor

                        # Pop rate. Actuators are more-or-less inputs so have no rate?
                        node_comp_params.pop("rate")
                        # Reassign converter in case a node provides the implementation for multiple actuators
                        obj_comp_params["processor"] = agnostic_processor

        # Get set of node we are required to launch
        dependencies = list(set(dependencies))

        # Verify that no dependency is an unlisted actuator node.
        not_selected = [cname for cname in agnostic["actuators"] if cname not in params["config"]["actuators"]]
        for cname in not_selected:
            try:
                for entry in specific["actuators"][cname]:
                    node_name, node_comp, node_cname = (entry["name"], entry["component"], entry["cname"])
                    msg = (
                        f'There appears to be a dependency on enginenode "{node_name}" for the implementation of '
                        f'engine "{state["engine_id"]}" for object "{name}" to work. However, enginenode "{node_name}" is '
                        f'directly tied to an unselected actuator "{cname}". '
                        "The actuator must be selected to resolve the graph."
                    )
                    assert node_name not in dependencies, msg
            except KeyError:
                # We pass here, because if cname is not selected, but also not implemented,
                # we are sure that there is no dependency.
                pass

        # Add relevant nodes
        nodes = {name: NodeSpec(params) for name, params in nodes.items() if name in dependencies}
        engine.objects[name].nodes = list(nodes.keys())
        graph.add([n for _, n in nodes.items()])

        # Connect tick to engine
        for node_name, n in nodes.items():
            if "tick" not in n.config.inputs:
                continue
            graph.connect(source=engine.outputs.tick, target=n.inputs.tick)
            for cname, o in n.outputs.items():
                mapping = dict(
                    delay=0.0,
                    window=0,
                    skip=False,
                    processor=None,
                    space=o.space.to_dict(),
                    address=None,
                )
                cname = f"{node_name}/{cname}"
                engine.config.inputs.append(cname)
                with engine.inputs as i:
                    i[cname] = mapping
                graph.connect(source=o, target=engine.inputs[cname], skip=True)
                break  # We only need 1 output per node.

        # Interconnect nodes
        for source, target in connects:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target
            if not (source_name in dependencies or target_name in dependencies):
                continue  # Continue if both source and target are not included as dependencies
            # If both source and target are included as dependencies
            if source_name in dependencies and target_name in dependencies:
                state["connects"].append([source, target])

        # Replace agnostic connections
        for s, t in deepcopy(state["connects"]):
            s_name, s_comp, s_cname = s
            t_name, t_comp, t_cname = t
            if name not in [s_name, t_name]:
                continue
            if t_comp == "actuators":
                assert t_cname in actuators, f"Actuator `{s_cname}` of object `{name}` has no implementation in this engine."
                s = graph.get_view(s[0], s[1:])
                t = graph.get_view(t[0], t[1:])
                window, delay, skip = t.window, t.delay, t.skip
                # Disconnect "agnostic" connection
                if s_name == "environment":
                    action = s_cname
                    s = None
                else:
                    action = None
                    s = s
                graph.disconnect(source=s, target=t, action=action)
                # Connect node with engine node
                for p in actuators[t_cname]:
                    _name, _comp, _cname = p["name"], p["component"], p["cname"]
                    new_target = getattr(nodes[_name], _comp)[_cname]
                    graph.connect(source=s, target=new_target, action=action, window=window, delay=delay, skip=skip)
            elif s_comp == "sensors":
                assert s_cname in sensors, f"Sensor `{s_cname}` of object `{name}` has no implementation in this engine."
                s = graph.get_view(s[0], s[1:])
                t = graph.get_view(t[0], t[1:])
                window, delay, skip = t.window, t.delay, t.skip
                # Disconnect "agnostic" connection
                if t_name == "environment":
                    observation = t_cname
                    t = None
                else:
                    observation = None
                    t = t
                graph.disconnect(source=s, target=t, observation=observation)
                # Connect node with engine node
                for p in sensors[s_cname]:
                    _name, _comp, _cname = p["name"], p["component"], p["cname"]
                    new_source = getattr(nodes[_name], _comp)[_cname]
                    graph.connect(source=new_source, target=t, observation=observation, window=window, delay=delay, skip=skip)
            elif s_comp == "states":
                s = graph.get_view(s[0], s[1:])
                t = graph.get_view(t[0], t[1:])
                graph.disconnect(source=s, target=t)
                t.address = "%s/%s/%s" % s()

        # Remove disconnected object from graph
        d = graph.remove(name)
        assert len(d) == 0, f"Object {name} not fully disconnected."



# RxEAGER
from eagerx.core.constants import process

# OTHER
import importlib
import subprocess
from time import sleep
from typing import List, Dict, Union, Any, TYPE_CHECKING
from functools import partial

if TYPE_CHECKING:
    from eagerx.core.entities import Backend
    from eagerx.core.specs import BaseNodeSpec


def get_launch_cmd(executable: str, bnd: "Backend", ns: str, name: str, external: bool = False):
    node_type, file = executable.split(":=")
    if "python" in node_type:
        if ".py" not in file:
            file = importlib.import_module(file).__file__
    if external:
        file = "/".join(["<path>", "<to>", "<package>"] + file.split("/")[-3:])

    cmd_args = [
        file,
        "--backend",
        f"{bnd.entity_id}",
        "--loglevel",
        f"{bnd.log_level}",
        "--env",
        f"{ns.split('/')[1]}",
        "--name",
        f"{name}",
    ]
    return cmd_args


def initialize_nodes(
    nodes: Union["BaseNodeSpec", List["BaseNodeSpec"]],
    process_id: int,
    ns: str,
    message_broker: Any,
    is_initialized: Dict,
    sp_nodes: Dict,
    launch_nodes: Dict,
    rxnode_cls: Any = None,
    node_args: Dict = None,
):
    if rxnode_cls is None:
        from eagerx.core.executable_node import RxNode

        rxnode_cls = RxNode

    # Convert to list
    if not isinstance(nodes, list):
        nodes = [nodes]

    from eagerx.core.specs import BaseNodeSpec

    assert all([isinstance(n, BaseNodeSpec) for n in nodes])

    bnd = message_broker.backend

    for node in nodes:
        name = node.config.name

        # Flag to check if node is initialized
        is_initialized[name] = False

        # Verify that node name is unique
        node_address = ns + "/" + name
        assert (node_address not in sp_nodes) and (node_address not in launch_nodes), (
            'Node "%s" already exists. Node names must be unique!' % name
        )

        # Block env until all nodes are initialized
        def initialized(msg, name):
            is_initialized[name] = True

        sub = message_broker.backend.Subscriber(node_address + "/initialized", "int64", partial(initialized, name=name))
        message_broker.subscribers.append(sub)

        # Initialize node
        if node.config.process == process_id:  # Initialize inside this process
            if node_args is None:
                node_args = dict()
            sp_nodes[node_address] = rxnode_cls(name=node_address, message_broker=message_broker, **node_args)
            sp_nodes[node_address].node_initialized()

        elif (
            node.config.process == process.NEW_PROCESS and process_id == process.ENVIRONMENT
        ):  # Only environment can launch new processes (as it is the main_thread)
            assert "executable" in node.config, (
                'No executable defined. Node "%s" can only be launched as a separate process if an executable is specified.'
                % name
            )
            cmd = get_launch_cmd(node.config.executable, bnd, ns, name, external=False)
            launch_nodes[node_address] = subprocess.Popen(cmd)
        elif node.config.process == process.EXTERNAL:
            cmd = get_launch_cmd(node.config.executable, bnd, ns, name, external=True)
            cmd_joined = " ".join(cmd).replace("\n", "\\n")
            message_broker.backend.loginfo(f'Launch node "{name}" externally with: python3 {cmd_joined}')
        # else: node is launched in another (already launched) node's process (e.g. engine process).


def wait_for_node_initialization(is_initialized, backend, wait_time=0.3):
    iter = 0

    # Wait for nodes to be initialized
    while True:
        if iter == 0:
            sleep(0.1)
        else:
            sleep(wait_time)
        iter += 1
        not_init = []
        for name, flag in is_initialized.items():
            if not flag:
                not_init.append(name)
        if len(not_init) > 0:
            backend.loginfo_once('Waiting for nodes "%s" to be initialized.' % (str(not_init)))
        else:
            break



import ctypes
import multiprocessing
from typing import Optional, List, Tuple
from threading import Event
from multiprocessing.sharedctypes import RawArray, Value
from multiprocessing import Event as MpEvent
import signal

import time
import sys

import numpy as np
import cv2

import eagerx
import eagerx.core.register as register
from eagerx.core.specs import NodeSpec
from eagerx.utils.utils import initialize_processor, Msg


class EnvNode(eagerx.Node):
    @classmethod
    def make(cls, rate=None, log_level=eagerx.log.WARN, color="yellow"):
        """EnvNode Spec"""
        spec = cls.get_specification()

        # Modify default node params
        spec.config.name = "environment"
        spec.config.rate = rate
        spec.config.process = eagerx.process.ENVIRONMENT
        spec.config.color = color
        spec.config.log_level = log_level
        spec.config.inputs = []
        spec.config.outputs = []
        spec.config.states = []
        return spec

    def initialize(self, spec: NodeSpec):
        # Define observation buffers
        self.observation_buffer = dict()
        for cname, i in self.inputs.items():
            if isinstance(i["space"], dict):
                i["space"] = eagerx.Space.from_dict(i["space"])
            assert i["space"] is not None, f"No space defined for observation {cname}."
            assert i[
                "space"
            ].is_fully_defined, f"The space for observation {cname} is not fully defined (low, high, shape, dtype)."
            window = i["window"]
            self.observation_buffer[cname] = {
                "msgs": None,
                "processor": i["processor"],
                "window": window,
                "space": i["space"],
            }

        # Synchronization event
        self.obs_event = Event()
        self.action_event = Event()
        self.must_reset = False

        # Graceful signal handler
        def signal_handler(sig, frame):
            self.backend.logdebug("SIGINT caught!")
            self.obs_event.set()
            self.action_event.set()
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)

        # Define action buffers
        self.action_buffer = dict()
        for cname, i in self.outputs.items():
            if isinstance(i["processor"], dict):
                from eagerx.core.specs import ProcessorSpec

                i["processor"] = initialize_processor(ProcessorSpec(i["processor"]))
            if isinstance(i["space"], dict):
                i["space"] = eagerx.Space.from_dict(i["space"])
            assert i["space"] is not None, f"No space defined for action {cname}."
            assert i["space"].is_fully_defined, f"The space for action {cname} is not fully defined (low, high, shape, dtype)."

            self.action_buffer[cname] = {"msg": None, "processor": i["processor"], "space": i["space"]}

    def reset(self):
        self.must_reset = False

        # Set all observation messages to None
        for _name, buffer in self.observation_buffer.items():
            buffer["msgs"] = None

        # Set all action messages to None
        for _name, buffer in self.action_buffer.items():
            buffer["msg"] = None

    def callback(self, t_n: float, **kwargs: Optional[Msg]):
        for name, i in kwargs.items():
            buffer = self.observation_buffer[name]
            window = buffer["window"]

            extra = window - len(i.msgs)
            if extra > 0:
                # Only happens when skip=True && window > 0
                if len(i.msgs) == 0:
                    initial_obs = np.empty(buffer["space"].shape)
                    if len(initial_obs.shape) > 0:
                        initial_obs[:] = np.NaN
                    else:
                        initial_obs = np.NaN
                    i.msgs.append(initial_obs)
                    extra -= 1  # Subtract, because we appended the initial_obs
                msgs = extra * [i.msgs[0]] + i.msgs
            else:
                msgs = i.msgs
            buffer["msgs"] = np.array(msgs)

        if not self.must_reset:
            self.action_event.clear()  # Clear action event, so that we can block after setting obs
            self.obs_event.set()  # Signal environment that observations have been set.
            try:
                # flag = self.action_event.wait(5)  # Wait for actions to be set.
                flag = self.action_event.wait()  # Wait for actions to be set.
                if not flag:
                    raise KeyboardInterrupt
            except (KeyboardInterrupt, SystemExit):
                self.backend.logdebug(f"[{self.ns_name}] KEYBOARD INTERRUPT")
                raise

        if self.must_reset:
            return None
        else:
            # Fill output_msg with buffered actions
            output_msgs = dict()
            for name, buffer in self.action_buffer.items():
                output_msgs[name] = buffer["msg"]
            return output_msgs

    def shutdown(self):
        self.obs_event.set()


class ObservationsNode(eagerx.Node):
    @classmethod
    def make(cls, rate=1, log_level=eagerx.log.WARN, color="yellow"):
        """ObservationsNode spec"""
        spec = cls.get_specification()

        # Modify default node params
        spec.config.name = "env/observations"
        spec.config.rate = rate
        spec.config.process = eagerx.process.ENVIRONMENT
        spec.config.color = color
        spec.config.log_level = log_level
        spec.config.inputs = ["actions_set"]
        spec.config.outputs = ["set"]
        spec.config.states = []

        # Pre-set address to mock an actual input
        spec.inputs.actions_set.address = "env/actions/outputs/set"
        return spec

    def initialize(self, spec: NodeSpec):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")

    def reset(self):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")

    @register.inputs(actions_set=eagerx.Space(dtype="int64"))
    @register.outputs(set=eagerx.Space(dtype="int64"))
    def callback(self, t_n: float, **kwargs: Optional[Msg]):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")


class ActionsNode(eagerx.Node):
    @classmethod
    def make(cls, rate=1, log_level=eagerx.log.WARN, color="yellow"):
        """ActionsNode spec"""
        spec = cls.get_specification()

        # Modify default node params
        spec.config.name = "env/actions"
        spec.config.rate = rate
        spec.config.process = eagerx.process.ENVIRONMENT
        spec.config.color = color
        spec.config.log_level = log_level
        spec.config.inputs = ["step"]
        spec.config.outputs = ["set"]
        spec.config.states = []

        # Pre-set address to mock an actual input
        spec.inputs.step.address = "env/supervisor/outputs/step"
        return spec

    def initialize(self, spec: NodeSpec):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")

    def reset(self):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")

    @register.inputs(observations_set=eagerx.Space(dtype="int64"), step=eagerx.Space(dtype="int64"))
    @register.outputs(set=eagerx.Space(dtype="int64"))
    def callback(self, t_n: float, **kwargs: Optional[Msg]):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")


class RenderNode(eagerx.Node):
    @classmethod
    def make(
        cls,
        rate,
        display=True,
        log_level=eagerx.log.WARN,
        color="grey",
        process=eagerx.process.NEW_PROCESS,
        encoding="bgr",
    ):
        """RenderNode spec"""
        spec = cls.get_specification()

        # Modify default node params
        spec.config.name = "env/render"
        spec.config.rate = rate
        assert process == eagerx.process.NEW_PROCESS, "Render requires process=NEW_PROCESS, because of cv2 limitations."
        spec.config.process = process
        spec.config.color = color
        spec.config.log_level = log_level
        spec.config.inputs = ["image"]
        spec.config.outputs = ["done"]
        spec.config.states = []

        # Modify custom params
        spec.config.display = display
        spec.config.encoding = encoding

        # Pre-set window
        spec.inputs.image.window = 1
        return spec

    def initialize(self, spec):
        self.display = spec.config.display
        self.window = None
        self.encoding = spec.config.encoding
        self.last_image = np.empty(shape=(0, 0, 3), dtype="uint8")
        self.render_toggle = Value("i", False)
        self.sub_toggle = self.backend.Subscriber("%s/%s/toggle" % (self.ns, self.name), "bool", self._set_render_toggle)
        self.sub_get = self.backend.Subscriber("%s/%s/get_last_image" % (self.ns, self.name), "bool", self._get_last_image)
        self.pub_set = self.backend.Publisher("%s/%s/set_last_image" % (self.ns, self.name), "uint8")

        # Setup async imshow (opening, closing, and imshow must all be in the same thread).
        assert hasattr(spec.inputs.image.space, "shape"), (
            "The node's outputs that is connected to " "the render node must have a space with shape defined."
        )
        self.img_event = MpEvent()
        self.p = None  # Set in _start_render_process(..)
        self.shared_array_np = None  # Set in _start_render_process(..)

    def _start_render_process(self, shape):
        size = None
        for i in shape:
            if size is None:
                size = i
            else:
                size *= i
        self.shared_array_np = np.ndarray(shape, dtype="uint8", buffer=RawArray(ctypes.c_uint8, size))
        args = (self.ns_name, shape, self.shared_array_np, self.img_event, self.render_toggle)
        self.p = multiprocessing.Process(target=self._async_imshow, args=args)
        self.p.start()

    @staticmethod
    def _async_imshow(ns_name: str, shape: Tuple[int], shared_array: np.ndarray, img_event: Value, render_toggle: Value):
        img = shared_array.view(dtype="uint8").reshape(*shape)
        try:
            while True:
                img_event.wait()  # Wait for event (ie new image or close window)
                img_event.clear()  # Clear event
                if render_toggle.value:
                    cv2.imshow(f"{ns_name}", img)
                cv2.waitKey(1)
        except (KeyboardInterrupt, SystemExit):
            # print(f"[{self.ns_name}]: KeyboardInterrupt caught")
            return

    def _set_render_toggle(self, msg):
        if msg:
            self.backend.logdebug("START RENDERING!")
        else:
            self.backend.logdebug("STOP RENDERING!")
        self.render_toggle.value = msg

    def _get_last_image(self, msg):
        self.pub_set.publish(self.last_image)

    def reset(self):
        pass

    @register.inputs(image=eagerx.Space(dtype="uint8"))
    @register.outputs(done=eagerx.Space(low=0, high=1, shape=(), dtype="int64"))
    def callback(self, t_n: float, image: Optional[Msg] = None):
        if image.msgs[-1].ndim >= 3:
            self.last_image = image.msgs[-1] if self.encoding == "rgb" else cv2.cvtColor(image.msgs[-1], cv2.COLOR_BGR2RGB)
            empty = len(image.msgs[-1]) == 0
            if not empty and self.display and self.render_toggle.value:
                img = image.msgs[-1] if self.encoding == "bgr" else cv2.cvtColor(image.msgs[-1], cv2.COLOR_RGB2BGR)
                if self.p is None:
                    self._start_render_process(img.shape)  # Start rendering in separate process
                np.copyto(self.shared_array_np, img)
            self.img_event.set()  # Signal async_imshow thread
        # Fill output_msg with 'done' output --> signals that we are done rendering
        output_msgs = dict(done=1)
        return output_msgs

    def shutdown(self):
        self.backend.logdebug(f"[{self.name}] {self.name}.shutdown() called.")
        self.sub_toggle.unregister()
        self.sub_get.unregister()
        self.pub_set.unregister()

        # Close render window
        if self.p is not None:
            self.p.kill()
        self.backend.logdebug(f"[{self.name}] {self.name}.shutdown() done.")


class ColabRender(eagerx.Node):
    @classmethod
    def make(
        cls,
        rate: int,
        process: int = eagerx.process.ENVIRONMENT,
        fps: int = 25,
        shape: Optional[List[int]] = None,
        maxlen: int = 200,
        subsample: bool = True,
        log_level: int = eagerx.log.WARN,
        color: str = "grey",
        encoding="bgr",
    ):
        """ColabRender spec"""
        spec = cls.get_specification()

        # Modify default node params
        spec.config.name = "env/render"
        spec.config.rate = rate
        assert process == eagerx.process.ENVIRONMENT, "ColabRendering must happen in the same process as the environment."
        spec.config.process = process
        spec.config.color = color
        spec.config.log_level = log_level
        spec.config.inputs = ["image"]
        spec.config.outputs = ["done"]
        spec.config.states = []

        # Custom params
        spec.config.encoding = encoding
        spec.config.fps = fps
        spec.config.shape = shape if isinstance(shape, list) else [64, 64]
        spec.config.maxlen = maxlen
        spec.config.subsample = True

        # Pre-set window
        spec.inputs.image.window = 1
        return spec

    def initialize(self, spec):
        # todo: Overwrite fps if higher than rate
        # todo: Subsample if fps lower than rate * real_time_factor
        # todo: set node_fps either slightly higher or lower than js_fps?
        # todo: Avoid overflowing buffer
        try:
            from eagerx.core.colab_render import InlineRender

            # Store cls as attribute so that it can be initialized in the callback
            self.window_cls = InlineRender
        except ImportError as e:
            self.backend.logerr(f"{e}. This node `ColabRender` can only be used in google colab.")
            raise
        self.dt_fps = 1 / spec.config.fps
        self.subsample = spec.config.subsample
        self.fps = spec.config.fps
        self.shape = spec.config.shape
        self.maxlen = spec.config.maxlen
        self.window = None
        self.encoding = spec.config.encoding
        self.last_image = np.empty(shape=(0, 0, 3), dtype="uint8")
        self.render_toggle = False
        self.sub_toggle = self.backend.Subscriber("%s/%s/toggle" % (self.ns, self.name), "bool", self._set_render_toggle)
        self.sub_get = self.backend.Subscriber("%s/%s/get_last_image" % (self.ns, self.name), "bool", self._get_last_image)
        self.pub_set = self.backend.Publisher("%s/%s/set_last_image" % (self.ns, self.name), "uint8")

    def _set_render_toggle(self, msg):
        if msg:
            self.backend.logdebug("START RENDERING!")
        else:
            self.backend.logdebug("STOP RENDERING!")
        self.render_toggle = msg

    def _get_last_image(self, msg):
        self.pub_set.publish(self.last_image)

    def reset(self):
        self.last_image = np.empty(shape=(0, 0, 3), dtype="uint8")

    @register.inputs(image=None)
    @register.outputs(done=eagerx.Space(low=0, high=1, shape=(), dtype="int64"))
    def callback(self, t_n: float, image: Optional[eagerx.utils.utils.Msg] = None):
        # Fill output_msg with 'done' output --> signals that we are done rendering
        output_msgs = dict(done=0)
        # Grab latest image
        self.last_image = image.msgs[-1] if self.encoding == "rgb" else cv2.cvtColor(image.msgs[-1], cv2.COLOR_BGR2RGB)
        # If too little time has passed, do not add frame (avoid buffer overflowing)
        if self.window is None:
            self.window = self.window_cls(fps=self.fps, maxlen=self.maxlen, shape=self.shape)
        elif not time.time() > (self.dt_fps + self.window.timestamp):
            return output_msgs
        # Check if frame is not empty
        empty = len(self.last_image.data) == 0
        if not empty and self.render_toggle:
            # Convert to rgb (from bgr)
            img = image.msgs[-1] if self.encoding == "rgb" else cv2.cvtColor(image.msgs[-1], cv2.COLOR_BGR2RGB)
            # Add image to buffer (where it is send async to javascript window)
            self.window.buffer_images(img)
        return output_msgs

    def shutdown(self):
        self.backend.logdebug(f"[{self.name}] {self.name}.shutdown() called.")
        self.sub_toggle.unregister()
        self.sub_get.unregister()
        self.pub_set.unregister()



# Rx imports
import eagerx.utils.utils
import eagerx.core.rx_message_broker
import eagerx.core.rx_operators
import eagerx.core.rx_pipelines
from eagerx.core.entities import BaseNode
from eagerx.utils.utils import (
    load,
    initialize_processor,
    get_param_with_blocking,
)
from eagerx.core.nodes import EnvNode
import eagerx

# OTHER
from threading import Event


class SupervisorNode(BaseNode):
    def __init__(self, env_node: EnvNode, **kwargs):
        super().__init__(**kwargs)

        self.subjects = None
        self.env_node = env_node

        # Render
        self.last_image = None
        self._image_event = Event()
        self.render_toggle = False
        self.pub_get_last_image = self.backend.Publisher(f"{self.ns}/env/render/get_last_image", "bool")
        self.sub_set_last_image = self.backend.Subscriber(
            f"{self.ns}/env/render/set_last_image", "uint8", self._last_image_callback
        )
        self.render_toggle_pub = self.backend.Publisher(f"{self.ns}/env/render/toggle", "bool")

        # Initialize buffer to hold desired reset states
        self.state_buffer = dict()
        for cname, i in self.states.items():
            if isinstance(i["processor"], dict):
                from eagerx.core.specs import ProcessorSpec

                i["processor"] = initialize_processor(ProcessorSpec(i["processor"]))
            if isinstance(i["space"], dict):
                i["space"] = eagerx.Space.from_dict(i["space"])
            assert i["space"] is not None, f"No space defined for state {cname}."
            assert i["space"].is_fully_defined, f"The space for state {cname} is not fully defined (low, high, shape, dtype)."
            self.state_buffer[cname] = {"msg": None, "processor": i["processor"], "space": i["space"]}

        # Required for reset
        self._step_counter = 0

    def _set_subjects(self, subjects):
        self.subjects = subjects

    def start_render(self):
        if not self.render_toggle:
            self.render_toggle = True
            self.render_toggle_pub.publish(self.render_toggle)

    def stop_render(self):
        if self.render_toggle:
            self.render_toggle = False
            self.render_toggle_pub.publish(self.render_toggle)

    def get_last_image(self):
        self._image_event.clear()
        self.pub_get_last_image.publish(True)
        self._image_event.wait()
        return self.last_image

    def _last_image_callback(self, msg):
        self.last_image = msg
        self._image_event.set()

    def _get_states(self, reset_msg):
        # Fill output_msg with buffered states
        msgs = dict()
        for name, buffer in self.state_buffer.items():
            if buffer["msg"] is None:
                msgs[name + "/done"] = True
            else:
                msgs[name + "/done"] = False
                msgs[name] = buffer["msg"]
                buffer["msg"] = None  # After sending state, set msg to None
        return msgs

    def reset(self):
        self.env_node.obs_event.clear()
        self.env_node.must_reset = True
        self.env_node.action_event.set()
        self.subjects["start_reset"].on_next(0)
        self._step_counter = 0
        try:
            flag = self.env_node.obs_event.wait()
            if not flag:
                raise KeyboardInterrupt
        except (KeyboardInterrupt, SystemExit):
            self.backend.logdebug("[reset] KEYBOARD INTERRUPT")
            raise
        self.backend.logdebug("FIRST OBS RECEIVED!")

    def step(self):
        self.env_node.obs_event.clear()
        self.env_node.action_event.set()
        self._step_counter += 1
        try:
            flag = self.env_node.obs_event.wait()
            if not flag:
                raise KeyboardInterrupt
        except (KeyboardInterrupt, SystemExit):
            self.backend.logdebug("[step] KEYBOARD INTERRUPT")
            raise
        self.backend.logdebug("STEP END")

    def shutdown(self):
        self.env_node.action_event.set()
        self.pub_get_last_image.unregister()
        self.sub_set_last_image.unregister()
        self.render_toggle_pub.unregister()


class Supervisor(object):
    def __init__(self, name, message_broker, env_node: EnvNode):
        self.name = name
        self.ns = "/".join(name.split("/")[:2])
        self.mb = message_broker
        self.backend = message_broker.backend
        self.initialized = False
        # self.sync = sync # todo: needed?
        self.has_shutdown = False

        # Prepare input & output topics
        outputs, states, self.node = self._prepare_io_topics(self.name, env_node)

        # Initialize reactive pipeline
        rx_objects, env_subjects = eagerx.core.rx_pipelines.init_supervisor(
            self.ns, self.node, outputs=outputs, state_outputs=states
        )
        self.node._set_subjects(env_subjects)
        self.mb.add_rx_objects(node_name=name, node=self, **rx_objects)

    def node_initialized(self):
        # Notify env that node is initialized
        self.init_pub = self.backend.Publisher(self.name + "/initialized", "int64")
        self.init_pub.publish(0)

        if not self.initialized:
            self.backend.loginfo('Node "%s" initialized.' % self.name)
        self.initialized = True

    def _prepare_io_topics(self, name: str, env_node: EnvNode):
        params = get_param_with_blocking(name, self.backend)

        # Get info from engine on reactive properties
        sync = get_param_with_blocking(self.ns + "/sync", self.backend)
        real_time_factor = get_param_with_blocking(self.ns + "/real_time_factor", self.backend)
        simulate_delays = get_param_with_blocking(self.ns + "/simulate_delays", self.backend)

        # Prepare output topics
        for i in params["outputs"]:
            if isinstance(i["processor"], dict):
                from eagerx.core.specs import ProcessorSpec

                i["processor"] = initialize_processor(ProcessorSpec(i["processor"]))
            if isinstance(i["space"], dict):
                i["space"] = eagerx.Space.from_dict(i["space"])

        # Prepare state topics
        for i in params["states"]:
            if isinstance(i["processor"], dict):
                from eagerx.core.specs import ProcessorSpec

                i["processor"] = initialize_processor(ProcessorSpec(i["processor"]))
            if isinstance(i["space"], dict):
                i["space"] = eagerx.Space.from_dict(i["space"])

        # Convert lists to dicts
        params["outputs"] = {i["name"]: i for i in params["outputs"]}
        params["states"] = {i["name"]: i for i in params["states"]}

        # Get node
        node_cls = load(params["node_type"])
        node = node_cls(
            ns=self.ns,
            message_broker=self.mb,
            sync=sync,
            real_time_factor=real_time_factor,
            simulate_delays=simulate_delays,
            env_node=env_node,
            params=params,
        )

        # Convert to tuple for reactive pipeline.
        outputs = tuple([value for key, value in params["outputs"].items()])
        states = tuple([value for key, value in params["states"].items()])

        return outputs, states, node

    def _shutdown(self):
        self.backend.logdebug(f"[{self.name}] Supervisor._shutdown() called.")
        self.init_pub.unregister()

    def node_shutdown(self):
        if not self.has_shutdown:
            self.backend.logdebug(f"[{self.name}] Supervisor.node_shutdown() called.")
            self.backend.loginfo(f"[{self.name}] Shutting down.")
            self._shutdown()
            self.node.shutdown()
            self.mb.shutdown()
            self.has_shutdown = True



import yaml
from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
import networkx as nx
from typing import List, Union, Dict, Tuple, Optional, Any
from eagerx.utils.utils import (
    is_compatible,
    supported_types,
)
from eagerx.utils.utils_sub import substitute_args
from eagerx.utils.network_utils import (
    episode_graph,
    plot_graph,
    color_nodes,
    color_edges,
    is_stale,
)
from eagerx.core.graph import merge
from eagerx.core.specs import NodeSpec, ProcessorSpec, EntitySpec
from eagerx.core.view import SpecView
import eagerx.core.log as log

yaml.Dumper.ignore_aliases = lambda *args: True


class EngineGraph:
    def __init__(self, state: Dict):
        self._state = state

    def __str__(self):
        return yaml.dump(self._state)

    @classmethod
    def create(cls, actuators: Optional[List[Dict]] = None, sensors: Optional[List[Dict]] = None):
        nodes = []

        from eagerx.core.entities import EngineNode

        # Create actuator node
        spec = EngineNode.pre_make(None, None)
        spec.config.name = "actuators"
        spec.config.color = "yellow"
        nodes.append(spec)
        for cname, params in actuators.items():
            # todo: Determine dtype, based on user-defined processor for target and source
            spec.add_output(
                cname,
                space=params.space.to_dict() if params.space is not None else None,
                processor=ProcessorSpec(params.processor.to_dict()) if params.processor is not None else None,
            )
            spec.add_input(
                cname,
                space=params.space.to_dict() if params.space is not None else None,
                skip=params.skip,
                window=params.window,
                processor=ProcessorSpec(params.processor.to_dict()) if params.processor is not None else None,
            )
            spec.config.outputs.append(cname)

        # Create sensor node
        spec = EngineNode.pre_make(None, None)
        spec.config.name = "sensors"
        spec.config.color = "yellow"
        nodes.append(spec)
        for cname, params in sensors.items():
            spec.config.inputs.append(cname)
            spec.add_input(
                cname,
                space=params.space.to_dict() if params.space is not None else None,
                processor=ProcessorSpec(params.processor.to_dict()) if params.processor is not None else None,
            )

        # Create a state
        state = dict(nodes=dict(), connects=list(), backup=dict(), gui_state=dict())
        graph = cls(state)
        graph.add(nodes)
        return graph

    def add(self, nodes: Union[NodeSpec, List[NodeSpec]]) -> None:
        """Add nodes to the graph.

        :param nodes: Nodes/objects to add.
        """
        if not isinstance(nodes, list):
            nodes = [nodes]

        for node in nodes:
            # Check spec
            self._check_spec(node)

            name = node.config.name
            assert name not in self._state["nodes"], (
                'There is already a node or object registered in this graph with name "%s".' % name
            )

            # Add node to state
            self._state["nodes"][name] = node._params
            self._state["backup"][name] = node.params

    def remove(self, names: Union[Union[str, EntitySpec], List[Union[str, EntitySpec]]]) -> None:
        """Removes a node from the graph.

        - First, all associated connections are disconnected.

        - Then, removes the nodes/objects.

        :param names: Either the name or spec of the node/object that is to be removed.
        """
        if not isinstance(names, list):
            names = [names]
        for name in names:
            if isinstance(name, EntitySpec):
                name = name.params["config"]["name"]
                assert name in self._state["nodes"], f" No entity with name '{name}' in graph."
            for source, target in deepcopy(self._state["connects"]):
                if name in [source[0], target[0]]:
                    if source[0] == "actuators":
                        actuator = source[2]
                        source = None
                    else:
                        actuator = None
                        source = self.get_view(source[0], source[1:])
                    if target[0] == "sensors":
                        sensor = target[2]
                        target = None
                    else:
                        sensor = None
                        target = self.get_view(target[0], target[1:])
                    self.disconnect(source, target, actuator, sensor)
            self._state["nodes"].pop(name)

    def add_component(self, entry: SpecView) -> None:
        """Selects an available component entry (e.g. input, output, etc...) that was not already selected.

        :param entry: Selects the entry, so that it can be connected.
        """
        name, component, cname = entry()
        # Add cname to selection list if it is not already selected
        params = self._state["nodes"][name]
        assert cname not in params["config"][component], f'"{cname}" already selected in "{name}" under {component}.'
        params["config"][component].append(cname)

    def remove_component(self, entry: SpecView) -> None:
        """Deselects a component entry (e.g. input, output, etc...) that was selected.

                - First, all associated connections are disconnected.

                - Then, deselects the component entry.

        :param entry: Deselects the entry.
        """
        self._is_selected(self._state, entry)

        # Disconnect component entry
        self._disconnect_component(entry)

        # Remove cname from selection list
        name, component, cname = entry()
        params = self._state["nodes"][name]
        params["config"][component].remove(cname)

    def connect(
        self,
        source: Optional[SpecView] = None,
        target: Optional[SpecView] = None,
        actuator: str = None,
        sensor: str = None,
        # processor: Optional[ProcessorSpec] = None,
        window: Optional[int] = None,
        delay: Optional[float] = None,
        skip: Optional[bool] = None,
    ) -> None:
        """Connect an actuator/source to a sensor/target.

        :param source: Compatible source type is :attr:`~eagerx.core.specs.NodeSpec.outputs`.
        :param target: Compatible target type is :attr:`~eagerx.core.specs.NodeSpec.inputs`.
        :param actuator: String name of the actuator.
        :param sensor: String name of the sensor.
        :param window: A non-negative number that specifies the number of messages to pass to the
                       node's :func:`~eagerx.core.entities.EngineNode.callback`.

                       - *window* = 1: Only the last received input message.

                       - *window* = *x* > 1: The trailing last *x* received input messages.

                       - *window* = 0: All input messages received since the last call to the
                         node's :func:`~eagerx.core.entities.EngineNode.callback`.

                       .. note:: With *window* = 0, the number of input messages may vary and can even be zero.
        :param delay: A non-negative simulated delay (seconds). This delay is ignored if
                      :attr:`~eagerx.core.entities.Engine.simulate_delays` = True
                      in the engine's :func:`~eagerx.core.entities.Engine.spec`.
        :param skip: Skip the dependency on this input during the first call to the node's :func:`~eagerx.core.entities.EngineNode.callback`.
                     May be necessary to ensure that the connected graph is directed and acyclic.
        """
        flag = not source or not actuator
        assert flag, (
            f'You cannot specify an actuator if you wish to connect actuator "{actuator}", '
            "as the actuator will act as the source."
        )
        flag = not target or not sensor
        assert flag, (
            f'You cannot specify a target if you wish to connect sensor "{sensor}", ' "as the sensor will act as the target."
        )
        assert not (actuator and sensor), "You cannot connect an actuator directly to a sensor."

        if actuator:  # source = actuator
            source = self.get_view("actuators", ["outputs", actuator])
            if target.processor is not None:
                msg = (
                    f'Cannot specify a processor for actuator "{actuator}", '
                    "because there is already one specified in the agnostic graph definition. "
                    "You can only have one processor."
                )
                assert source.processor is None, msg
            assert window is None, (
                f'Cannot specify a window when connecting actuator "{actuator}". '
                f"You can only do that in the agnostic object definition."
            )
            assert delay is None, (
                f'Cannot specify a delay when connecting actuator "{actuator}". '
                f"You can only do that in the agnostic object definition."
            )
            assert skip is None, (
                f'Cannot specify a skip when connecting actuator "{actuator}". '
                f"You can only do that in the agnostic object definition."
            )
        elif sensor:  # target = sensor
            target = self.get_view("sensors", ["inputs", sensor])
            # Check for processor clash.
            # I.e. if both agnostic sensor & enginenode output have a processor.
            if source.processor is not None:
                src_name, src_comp, src_cname = source()
                msg = (
                    "Processor clash detected! "
                    f"You attempt to connect EngineNode output '{src_name}.{src_comp}.{src_cname}' to sensor '{sensor}'. "
                    f"There is a processor defined for '{src_name}.{src_comp}.{src_cname}', but sensor '{sensor}' also has a "
                    f"a processor. Only one can be used."
                )
                assert target.processor is None, msg
            assert window is None, (
                f'Cannot specify a window when connecting sensor "{sensor}".'
                " You can only do that in the agnostic object definition."
            )
            assert delay is None, (
                f'Cannot specify a delay when connecting sensor "{sensor}".'
                " You can only do that in the agnostic object definition."
            )
            assert skip is None, (
                f'Cannot specify a skip when connecting sensor "{sensor}".'
                " You can only do that in the agnostic object definition."
            )
        self._connect(source, target, window, delay, skip)

    def _connect(
        self,
        source: Optional[SpecView] = None,
        target: Optional[SpecView] = None,
        window: Optional[int] = None,
        delay: Optional[float] = None,
        skip: Optional[bool] = None,
    ):
        """
        Method to connect a source to a target. For actuators/sensors, first a (new) disconnected entry must be created,
        after which an additional call to connect_actuator/sensor is required before calling this method.
        For more info, see self.connect.
        """
        # Perform checks on source
        self._is_selected(self._state, source)

        # Perform checks on target
        self._is_selected(self._state, target)

        # Make sure that source and target are compatible
        self._is_compatible(self._state, source, target)

        # Make sure that target is not already connected.
        target_name, target_comp, target_cname = target()
        for s, t in self._state["connects"]:
            t_name, t_comp, t_cname = t
            flag = not (target_name == t_name and target_comp == t_comp and target_cname == t_cname)
            assert flag, f'Target "{target}" is already connected to source "{s}"'

        # Add properties to target params
        if window is not None:
            self.set({"window": window}, target)
        if delay is not None:
            self.set({"delay": delay}, target)
        if skip is not None:
            self.set({"skip": skip}, target)

        # Add connection
        connect = [list(source()), list(target())]
        self._state["connects"].append(connect)

    def disconnect(
        self,
        source: Optional[SpecView] = None,
        target: Optional[SpecView] = None,
        actuator: str = None,
        sensor: str = None,
    ) -> None:
        """Disconnect an actuator/source from a sensor/target.

        :param source: Compatible source type is :attr:`~eagerx.core.specs.NodeSpec.outputs`.
        :param target: Compatible target type is :attr:`~eagerx.core.specs.NodeSpec.inputs`.
        :param actuator: String name of the actuator.
        :param sensor: String name of the sensor.
        """
        assert not source or not actuator, (
            "You cannot specify a source if you wish to disconnect actuator ",
            f'"{actuator}", as the actuator will act as the source.',
        )
        assert not target or not sensor, (
            "You cannot specify a target if you wish to disconnect sensor "
            f'"{sensor}", as the sensor will act as the target.'
        )
        assert not (sensor and actuator), "You cannot disconnect an actuator from a sensor, as such a connection cannot exist."

        # Create source & target entries
        if actuator:
            source = self.get_view("actuators", ["outputs", actuator])
        if sensor:
            target = self.get_view("sensors", ["inputs", sensor])

        self._disconnect(source, target)

    def _disconnect(self, source: SpecView, target: SpecView):
        """
        Disconnects a source from a target. The target is reset in self._state to its disconnected state.
        """
        # Check if connection exists
        self._is_selected(self._state, target)

        # Check if connection exists
        self._is_selected(self._state, source)
        connect_exists = False
        idx_connect = None
        for idx, c in enumerate(self._state["connects"]):
            if list(source()) == c[0] and list(target()) == c[1]:
                connect_exists = True
                idx_connect = idx
                break
        assert connect_exists, (
            f"The connection with source={source()} and target={target()} cannot be removed," " because it does not exist."
        )

        # Pop the connection from the state
        self._state["connects"].pop(idx_connect)

        # Reset target params to disconnected state (reset to go back to default yaml), i.e. reset window/delay/skip.
        target_name, target_comp, target_cname = target()
        target_params = self._state["nodes"][target_name]
        target_params[target_comp][target_cname] = self._state["backup"][target_name][target_comp][target_cname]

    def _disconnect_component(self, entry: SpecView):
        """
        Disconnects all associated connects from self._state.
        """
        was_connected = False
        name, component, cname = entry()
        for source, target in deepcopy(self._state["connects"]):
            source = self.get_view(source[0], source[1:])
            target = self.get_view(target[0], target[1:])
            self._is_selected(self._state, source)
            self._is_selected(self._state, target)
            source_name, source_comp, source_cname = source()
            target_name, target_comp, target_cname = target()
            if name == source_name and component == source_comp and cname == source_cname:
                self.disconnect(source, target)
                was_connected = True
            elif name == target_name and component == target_comp and cname == target_cname:
                self.disconnect(source, target)
                was_connected = True
        return was_connected

    def set(self, mapping: Any, entry: Optional[SpecView], parameter: Optional[str] = None) -> None:
        """Sets the parameters of a node.

        :param mapping: Either a mapping with *key* = *parameter*,
                        or a single value that corresponds to the optional *parameter* argument.
        :param entry: The entry whose parameters are mutated.
        :param parameter: If only a single value needs to be set. See documentation for *mapping*.
        """
        assert not entry()[0] == "actuators", (
            "Cannot change the actuator parameters here, "
            "in an engine specific implementation. That is only possible in the "
            "object's agnostic definition."
        )
        assert not entry()[0] == "sensors", (
            "Cannot change the sensor parameters here, "
            "in an engine specific implementation. That is only possible in the "
            "object's agnostic definition."
        )

        if parameter is None:
            if isinstance(mapping, SpecView):
                mapping = mapping.to_dict()
            assert isinstance(mapping, dict), "Can only set mappings of type dict. Else, also set 'parameter=<param_name>'."
        else:
            mapping = {parameter: mapping}
        for parameter, value in mapping.items():
            if parameter:
                getattr(entry, parameter)  # Check if parameter exists
            if parameter == "processor":
                msg = (
                    "Skipping processor. Cannot change the processor with this method. "
                    "Add output processors before connecting, and input processors when making a connection."
                )
                log.logwarn(msg)
            else:
                t = entry()
                name = t[0]
                if t[1] == "config":
                    assert parameter not in [
                        "sensors",
                        "actuators",
                        "targets",
                        "states",
                        "inputs",
                        "outputs",
                    ], "You cannot modify component parameters with this function. Use _add/remove_component(..) instead."
                    assert parameter not in ["name"], f"You cannot rename '{name}'."
                    p = self._state["nodes"][name]["config"]
                else:
                    name, component, cname = entry()
                    p = self._state["nodes"][name][component][cname]
                self._set(p, {parameter: value})

    def get(
        self,
        entry: Optional[Union[SpecView, EntitySpec]] = None,
        actuator: Optional[str] = None,
        sensor: Optional[str] = None,
        parameter: Optional[str] = None,
    ) -> Any:
        """Fetches the parameters of a node/actuator/sensor.

        :param entry: The entry whose parameters are fetched.
        :param actuator: Actuator name whose parameters are fetched.
        :param sensor: Sensor name whose parameters are fetched.
        :param parameter: If only a single parameter needs to be fetched.
        :return: Parameters
        """
        if isinstance(entry, EntitySpec):
            name = entry.params["config"]["name"]
            assert name in self._state["nodes"], f" No entity with name '{name}' in graph."
            return self._state["nodes"][name]
        self._correct_signature(entry, actuator, sensor)
        if actuator:
            entry = self.get_view("actuators", ["outputs", actuator])
        if sensor:
            entry = self.get_view("sensors", ["inputs", sensor])
        if parameter:
            return getattr(entry, parameter)
        else:
            return entry

    def get_spec(self, name: str) -> NodeSpec:
        """Get Spec from the graph

        :param name: Name
        :return: The specification of the entity.
        """
        assert name in self._state["nodes"], f" No entity with name '{name}' in graph."
        params = self._state["nodes"][name]
        spec = NodeSpec(params)
        return spec

    def _node_dependencies(self, state):
        import networkx as nx

        state = deepcopy(state)
        G = self._generate_graph(state)
        G_rev = G.reverse(copy=True)

        # fig_env, ax_env = plt.subplots(nrows=1, ncols=1)
        # ax_env.set_title('Engine-specific graph')
        # _, _, _, pos = plot_graph(G, k=2, ax=ax_env)
        # plt.show()
        #
        # fig_env, ax_env = plt.subplots(nrows=1, ncols=1)
        # ax_env.set_title('Engine-specific graph (reversed)')
        # _, _, _, _ = plot_graph(G_rev, k=2, ax=ax_env, pos=pos)
        # plt.show()

        # check if node is synced
        def is_synced(node_name):
            return True if "tick" in state["nodes"][node_name]["config"]["inputs"] else False

        # Determine sensor dependencies
        dependencies = dict(sensors=dict(), actuators=dict())
        for cname in state["nodes"]["sensors"]["inputs"]:
            dependencies["sensors"][cname] = []
            target_name = f"sensors/{cname}"
            descendants = nx.descendants(G_rev, target_name)
            for source in descendants:
                node_name, source_cname = source.split("/")
                if node_name in ["actuators", "sensors"]:
                    continue
                dependencies["sensors"][cname].append(node_name)

        # Determine actuator dependency
        for cname in state["nodes"]["actuators"]["outputs"]:
            dependencies["actuators"][cname] = []
            source_name = f"actuators/{cname}"
            descendants = nx.descendants(G, source_name)
            for target in list(descendants):
                node_name, target_cname = target.split("/")
                if node_name in ["actuators", "sensors"]:
                    continue
                sync = is_synced(node_name)
                if sync:  # Only add downstream dependency if it is synced with the engine
                    dependencies["actuators"][cname].append(node_name)
                else:
                    # Pop from descendants list so that we only add reversed dependencies of synced engine nodes
                    # Node might still be added later-on as a dependency, if synced node requires output of async node.
                    descendants.remove(target)
            # Also add dependencies of all targets
            for target in descendants:
                nx.descendants(G_rev, target)
                for source in descendants:
                    node_name, source_cname = source.split("/")
                    if node_name in ["actuators", "sensors"]:
                        continue
                    dependencies["actuators"][cname].append(node_name)
            # Always add all nodes that are directly connected to actuator (even if they do not have an output)
            for source, target in state["connects"]:
                source_name, source_comp, source_cname = source
                target_name, target_comp, target_cname = target
                if source_name == "actuators" and cname == source_cname:
                    dependencies["actuators"][cname].append(target_name)
            dependencies["actuators"][cname] = list(set(dependencies["actuators"].pop(cname)))
        return dependencies

    def register(self):
        """Returns the nodes that make up this subgraph,
        and their relation to the registered actuators and sensors."""

        # Check if valid graph.
        assert self.is_valid(plot=False), "Graph not valid."

        # Find dependencies
        dependencies = self._node_dependencies(self._state)

        # Copy state
        state = deepcopy(self._state)

        # Add addresses based on connections
        actuators = dict()
        sensors = dict()
        connects = state["connects"]
        for source, target in connects:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target
            if source_name == "actuators":
                dependency = [f"$(ns obj_name)/{d}" for d in dependencies["actuators"][source_cname]]
                if source_cname not in actuators:
                    actuators[source_cname] = []
                entry = {
                    "name": f"$(ns obj_name)/{target_name}",
                    "component": target_comp,
                    "cname": target_cname,
                    "dependency": dependency,
                }
                actuators[source_cname].append(entry)
                continue  # we continue here, because the address for actuators is determined by an output from the agnostic graph.
            if target_name == "sensors":
                dependency = [f"$(ns obj_name)/{d}" for d in dependencies["sensors"][target_cname]]
                if target_cname not in sensors:
                    sensors[target_cname] = []
                entry = {
                    "name": f"$(ns obj_name)/{source_name}",
                    "component": source_comp,
                    "cname": source_cname,
                    "dependency": dependency,
                }
                sensors[target_cname].append(entry)
                continue  # we continue here, because the address for actuators is determined by an output from the agnostic graph.
            # state["nodes"][target_name][target_comp][target_cname]["address"] = address

            # Determine if dtypes match
            s = self.get_view(source[0], source[1:])
            t = self.get_view(target[0], target[1:])
            self._is_compatible(state, s, t)
            source_dtype = state["nodes"][source_name][source_comp][source_cname]["space"]["dtype"]
            state["nodes"][target_name][target_comp][target_cname]["dtype"] = source_dtype

        # Prepend object name to source & target name
        for source, target in connects:
            if source[0] != "actuators":
                source[0] = f"$(ns obj_name)/{source[0]}"
            if target[0] != "sensors":
                target[0] = f"$(ns obj_name)/{target[0]}"

        # Initialize param objects
        nodes = dict()
        from eagerx.core.specs import NodeSpec

        for name, params in state["nodes"].items():
            if "node_type" in params:
                if name == "actuators":
                    pass
                elif name == "sensors":
                    pass
                else:
                    spec = NodeSpec(params)
                    flag = True if "tick" not in spec.config.inputs else len(spec.config.outputs) > 0
                    assert flag, (
                        f"If Node `{spec.config.name}` must run synchronized with the Engine, "
                        f"it must have at least 1 (dummy?) output."
                    )

                    # Put node name into object namespace
                    name = f"$(ns obj_name)/{spec.config.name}"
                    spec.config.name = name
                    params = spec.params

                    # Substitute placeholder args of simnode
                    context = {"ns": {"node_name": name}, "config": params["config"]}
                    substitute_args(params, context, only=["config", "ns"])
                    nodes[name] = params

        return nodes, actuators, sensors, connects

    def gui(
        self, interactive: Optional[bool] = True, resolution: Optional[List[int]] = None, filename: Optional[str] = None
    ) -> Union[None, np.ndarray]:
        """Opens a graphical user interface of the graph.

        .. note:: Requires `eagerx-gui`:

                .. highlight:: python
                .. code-block:: python

                    pip3 install eagerx-gui

        :param interactive: If `True`, an interactive application is launched.
                            Otherwise, an RGB render of the GUI is returned.
                            This could be useful when using a headless machine.
        :param resolution: Specifies the resolution of the returned render when `interactive` is `False`.
                           If `interactive` is `True`, this argument is ignored.
        :param filename: If provided, the GUI is rendered to an svg file with this name.
                         If `interactive` is `True`, this argument is ignored.
        :return: RGB render of the GUI if `interactive` is `False`.
        """
        try:
            from eagerx_gui import launch_gui, render_gui
        except ImportError as e:
            log.logwarn(
                f"{e}. You will likely have to install eagerx-gui. It can be installed by running: pip install eagerx-gui"
            )
            return

        if interactive:
            self._state = launch_gui(deepcopy(self._state), is_engine=True)
        else:
            return render_gui(deepcopy(self._state), resolution=resolution, filename=filename, is_engine=True)

    @staticmethod
    def _get_address(source: Tuple[str, str, str], target: Tuple[str, str, str]):
        """Determine the address."""
        source_name, source_comp, source_cname = source
        target_name, target_comp, target_cname = target
        if source_name == "actuators":
            assert (
                not target_name == "sensors"
            ), f'A direct connection between a sensor "{target_cname}" and actuator "{source_cname}" cannot exist.'
            address = f"$(ns obj_name)/{source_name}/{source_cname}"
        elif target_name == "sensors":
            assert (
                not source_name == "actuators"
            ), f'A direct connection between a sensor "{target_cname}" and actuator "{source_cname}" cannot exist.'
            address = f"$(ns obj_name)/{target_name}/{target_cname}"
        else:
            address = f"$(ns obj_name)/{source_name}/{source_comp}/{source_cname}"
        return address

    @staticmethod
    def _is_selected(state: Dict, entry: SpecView):
        """
        Check if provided entry was selected in params.
        """
        name, component, cname = entry()
        params = state["nodes"][name]
        component = "outputs" if component == "feedthroughs" else component
        assert cname in params["config"][component], f'"{cname}" not selected in "{name}" under "config" in {component}.'

    @staticmethod
    def _is_compatible(state: Dict, source: SpecView, target: SpecView):
        """Check if provided the provided entries are compatible."""
        # Valid source and target components
        targets = ["inputs"]
        sources = ["outputs"]

        # Provided entries
        target_name, target_component, target_cname = target()
        source_name, source_component, source_cname = source()
        base_msg = (
            f"'{target_name}.{target_component}.{target_cname}' cannot be connected with "
            f"'{source_name}.{source_component}.{source_cname}')."
        )

        # Check if target & source are validly chosen
        assert target_component in targets, f"{base_msg} '{target_component}' cannot be a target."
        assert source_component in sources, f"{base_msg} '{source_component}' cannot be a source."

        # Check if combinations are valid.
        if source_component in ["outputs", "sensors"]:
            valid = ["inputs", "feedthroughs", "actuators"]
            msg = f"{base_msg} '{source_component}' can only be connected to any of the components in '{valid}'."
            assert target_component in valid, msg
        else:  # source_component == "states":
            valid = ["targets"]
            msg = f"{base_msg} '{source_component}' can only be connected to any of the components in '{valid}'."
            assert target_component in valid, msg

            # Check that the state corresponds to an object (i.e. is an engine state)
            msg = (
                f"{base_msg} Only '{source_component}' of Objects can be connected to targets. '{source_name}' "
                "is not of type Object."
            )
            assert "node_type" not in state["nodes"][source_name], msg

        # Check if dtypes match
        source_params = state["nodes"][source_name][source_component][source_cname]
        target_params = state["nodes"][target_name][target_component][target_cname]
        assert (
            source_params["space"] is not None
        ), f"source={source_name}.{source_component}.{source_cname} does not have a space defined."
        assert (
            target_params["space"] is not None
        ), f"target={target_name}.{target_component}.{target_cname} does not have a space defined."
        source_dtype = source_params["space"]["dtype"]
        target_dtype = target_params["space"]["dtype"]
        try:
            is_compatible(source_dtype, target_dtype)
        except AssertionError as e:
            msg = (
                f"Incorrect connection of (source={source_name}.{source_component}.{source_cname}) with "
                f"(target={target_name}.{target_component}.{target_cname}): {e}"
            )
            raise IOError(msg)

    @staticmethod
    def _correct_signature(
        entry: Optional[SpecView] = None,
        actuator: Optional[str] = None,
        sensor: Optional[str] = None,
    ):
        # assert only action, only observation, or only entry
        if entry:  # component parameter
            assert actuator is None, "If 'entry' is specified, actuator argument cannot be specified."
            assert sensor is None, "If 'entry' is specified, sensor argument cannot be specified."
        if actuator:
            assert sensor is None, "If actuator is specified, sensor must be None."
            assert entry is None, "If actuator is specified, the 'entry' argument cannot be specified."
        if sensor:
            assert actuator is None, "If sensor is specified, action must be None."
            assert entry is None, "If actuator is specified, the 'entry' argument cannot be specified."

    def is_valid(self, plot=True) -> bool:
        """Checks the validity of the graph.

        - Checks if all selected
          :attr:`~eagerx.core.specs.NodeSpec.inputs` are connected.

        - Checks if the graph is directed and acyclic.

        :param plot: Flag to plot the graph. Can be helpful to identify cycles in the graph that break the required acyclic property.
        :returns: flag that specifies the validity of the graph.
        """
        return self._is_valid(self._state, plot=plot)

    @staticmethod
    def _is_valid(state, plot=True):
        state = deepcopy(state)
        EngineGraph.check_inputs_have_address(state)
        EngineGraph.check_graph_is_acyclic(state, plot=plot)
        return True

    @staticmethod
    def check_inputs_have_address(state):
        state = deepcopy(state)
        for source, target in state["connects"]:
            address = EngineGraph._get_address(source, target)
            target_name, target_comp, target_cname = target
            state["nodes"][target_name][target_comp][target_cname]["address"] = address

        for name, params in state["nodes"].items():
            for component in params["config"]:
                if component not in [
                    "inputs",
                    "outputs",
                    "targets",
                    "feedthroughs",
                    "states",
                ]:
                    continue
                for cname in params["config"][component]:
                    flag = cname in params[component]
                    assert flag, f'"{cname}" was selected in {component} of "{name}", but has no implementation.'
                    if component not in ["inputs", "targets", "feedthroughs"]:
                        continue
                    if name in ["sensors", "actuators"]:
                        continue
                    # Skip if tick
                    if cname == "tick" and component == "inputs":
                        continue
                    flag = params[component][cname]["address"] is not None
                    assert flag, (
                        f'"{cname}" was selected in {component} of "{name}", but no address was specified. '
                        f"Either deselect it, or connect it."
                    )
        return True

    @staticmethod
    def check_graph_is_acyclic(state, plot=True):
        # Generate graph
        G = EngineGraph._generate_graph(state)

        # Color nodes based on in/out going edges
        not_active = is_stale(G, exclude_skip=True)
        color_nodes(G)
        color_edges(G)

        # Check if graph is acyclic (excluding 'skip' edges)
        H, cycles = episode_graph(G)
        is_dag = nx.is_directed_acyclic_graph(H)

        # Plot graphs
        if plot:
            fig_env, ax_env = plt.subplots(nrows=1, ncols=1)
            ax_env.set_title("Engine-specific graph")
            _, _, _, pos = plot_graph(G, k=2, ax=ax_env)
            plt.show()

        # Assert if graph is a directed-acyclical graph (DAG)
        cycle_strs = ["Circular loops detected: "]
        for idx, connect in enumerate(cycles):
            connect.append(connect[0])
            s = " Loop %s: " % idx
            n = "\n" + "".join([" "] * len(s)) + "...-->"
            s = "\n\n" + s + "...-->"
            for idx in range(len(connect) - 1):
                tmp, target = connect[idx]
                source, tmp2 = connect[idx + 1]
                source_name, source_comp, source_cname = source
                target_name, target_comp, target_cname = target
                assert source_name == target_name, "Source and target not equal: %s, %s" % (source, target)
                connect_name = "%s/%s/%s][%s/%s/%s" % tuple(list(source) + list(target))
                node_name = ("Node: " + source_name).center(len(connect_name), " ")
                s += "[%s]-->" % connect_name
                n += "[%s]-->" % node_name
            s += "..."
            n += "..."
            cycle_strs.append(s)
            cycle_strs.append(n)
            connect.pop(-1)
        assert is_dag, "".join(cycle_strs)

        not_active_excl_sensors = [n for n in not_active if not n.split("/")[0] == "sensors"]
        assert len(not_active_excl_sensors) == 0, (
            'Stale episode graph detected. Nodes "%s" will be stale, while they must be active (i.e. connected) in order for the graph to resolve (i.e. not deadlock).'
            % not_active
        )
        return True

    @staticmethod
    def _generate_graph(state):
        # Add nodes
        G = nx.MultiDiGraph()
        # label_mapping = {'env/observations/set': 'observations', 'env/render/done': 'render'}
        for node, params in state["nodes"].items():
            default = params["config"]
            if "outputs" in default:
                has_tick = True if "tick" in default["inputs"] else False
                for cname in default["outputs"]:
                    name = "%s/%s" % (node, cname)
                    remain_active = True if node == "actuators" else False
                    always_active = True if node == "actuators" else False
                    G.add_node(
                        name,
                        remain_active=remain_active,
                        always_active=always_active,
                        is_stale=False,
                        has_tick=has_tick,
                    )
            if node == "sensors":
                for cname in default["inputs"]:
                    name = "%s/%s" % (node, cname)
                    G.add_node(
                        name,
                        remain_active=True,
                        always_active=False,
                        is_stale=False,
                        has_tick=False,
                    )

        # Add edges
        target_comps = ["inputs"]
        source_comps = ["outputs"]
        for source, target in state["connects"]:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target
            if source_comp in source_comps and target_comp in target_comps:
                source_edge = "%s/%s" % (source_name, source_cname)
                # Determine target node name
                target_edges = []
                target_default = state["nodes"][target_name]["config"]
                for cname in target_default["outputs"]:
                    target_edge = "%s/%s" % (target_name, cname)
                    target_edges.append(target_edge)
                if target_name == "sensors":
                    target_edge = "%s/%s" % (target_name, target_cname)
                    target_edges.append(target_edge)

                # Determine edges that do not break DAG property (i.e. edges that are skipped)
                if source_name == "actuators":
                    skip = state["nodes"][source_name]["inputs"][source_cname]["skip"]
                else:
                    skip = state["nodes"][target_name][target_comp][target_cname]["skip"]

                # Determine color
                color = "green" if skip else "black"
                style = "dotted" if skip else "solid"

                # Add edge
                for target_edge in target_edges:
                    G.add_edge(
                        source_edge,
                        target_edge,
                        color=color,
                        feedthrough=False,
                        style=style,
                        alpha=1.0,
                        is_stale=False,
                        skip=skip,
                        source=source,
                        target=target,
                    )

        # Color nodes based on in/out going edges
        # not_active = is_stale(G, exclude_skip=True)
        color_nodes(G)
        color_edges(G)
        return G

    @staticmethod
    def _get_view(spec, name: str, depth: Optional[List[str]] = None):
        depth = depth if depth else []
        # depth = [name] + depth
        return SpecView(spec, depth=depth, name=name)

    def get_view(self, name: str, depth: Optional[List[str]] = None):
        return self._get_view(self.get_spec(name), name, depth)

    @staticmethod
    @supported_types(str, int, list, float, bool, dict, EntitySpec, SpecView, None)
    def _set(state, mapping):
        merge(state, mapping)

    @staticmethod
    def _check_spec(spec):
        if spec.config.name in ["sensors", "actuators"]:
            return
        from eagerx.core.entities import EngineNode

        EngineNode.check_spec(spec)



#!/usr/bin/env python3

# import os
# if bool(eval(os.environ.get("EAGERX_COLAB", "0"))):
#     import site
#
#     site.addsitedir("/opt/ros/melodic/lib/python2.7/dist-packages")
#     site.addsitedir("/usr/lib/python2.7/dist-packages")

# Rx imports
from eagerx.core.entities import Backend
import eagerx.core.rx_message_broker
import eagerx.core.rx_operators
import eagerx.core.rx_pipelines
from eagerx.utils.utils import (
    load,
    initialize_processor,
    get_param_with_blocking,
)
from eagerx.core.executable_node import RxNode
from eagerx.utils.node_utils import wait_for_node_initialization

# Other imports
from threading import Condition
import argparse


class RxEngine(object):
    def __init__(self, name, message_broker):
        self.name = name
        self.ns = "/".join(name.split("/")[:2])
        self.mb = message_broker
        self.backend = message_broker.backend
        self.initialized = False
        self.has_shutdown = False

        # Prepare input & output topics
        (
            rate,
            inputs,
            outputs,
            states,
            engine_states,
            node_names,
            target_addresses,
            self.engine,
        ) = self._prepare_io_topics(self.name)

        # Initialize reactive pipeline
        rx_objects = eagerx.core.rx_pipelines.init_engine(
            self.ns,
            rate,
            self.engine,
            inputs,
            outputs,
            states,
            engine_states,
            node_names,
            target_addresses,
            self.mb,
        )
        self.mb.add_rx_objects(node_name=name, node=self, **rx_objects)
        self.mb.connect_io()
        self.cond_reg = Condition()

        # Prepare closing routine
        # self.backend.on_shutdown(self.node_shutdown)

    def node_initialized(self):
        with self.cond_reg:
            # Wait for all nodes to be initialized
            wait_for_node_initialization(self.engine.is_initialized, self.backend)

            # Notify env that node is initialized
            if not self.initialized:
                self.init_pub = self.backend.Publisher(self.name + "/initialized", "int64")
                self.init_pub.publish(0)
                self.backend.loginfo('Node "%s" initialized.' % self.name)
                self.initialized = True

    def _prepare_io_topics(self, name):
        params = get_param_with_blocking(name, self.backend)
        node_names = params["config"]["node_names"]
        target_addresses = params["config"]["target_addresses"]
        rate = params["config"]["rate"]

        # Get info on run-time settings
        sync = get_param_with_blocking(self.ns + "/sync", self.backend)
        real_time_factor = get_param_with_blocking(self.ns + "/real_time_factor", self.backend)
        simulate_delays = get_param_with_blocking(self.ns + "/simulate_delays", self.backend)

        # Prepare input topics
        for i in params["inputs"]:
            if isinstance(i["processor"], dict):
                from eagerx.core.specs import ProcessorSpec

                i["processor"] = initialize_processor(ProcessorSpec(i["processor"]))
            if isinstance(i["space"], dict):
                i["space"] = eagerx.Space.from_dict(i["space"])

        # Prepare output topics
        for i in params["outputs"]:
            if isinstance(i["processor"], dict):
                from eagerx.core.specs import ProcessorSpec

                i["processor"] = initialize_processor(ProcessorSpec(i["processor"]))
            if isinstance(i["space"], dict):
                i["space"] = eagerx.Space.from_dict(i["space"])

        # Prepare state topics
        for i in params["states"]:
            if isinstance(i["processor"], dict):
                from eagerx.core.specs import ProcessorSpec

                i["processor"] = initialize_processor(ProcessorSpec(i["processor"]))
            if isinstance(i["space"], dict):
                i["space"] = eagerx.Space.from_dict(i["space"])

        # Convert lists to dicts
        params["inputs"] = {i["name"]: i for i in params["inputs"]}
        params["outputs"] = {i["name"]: i for i in params["outputs"]}
        params["states"] = {i["name"]: i for i in params["states"]}

        # Get node
        node_cls = load(params["node_type"])
        node = node_cls(
            ns=self.ns,
            message_broker=self.mb,
            sync=sync,
            real_time_factor=real_time_factor,
            simulate_delays=simulate_delays,
            params=params,
            target_addresses=target_addresses,
            node_names=node_names,
        )

        # Convert to tuple for reactive pipeline.
        inputs = tuple([value for key, value in params["inputs"].items()])
        outputs = tuple([value for key, value in params["outputs"].items()])
        states = tuple([value for key, value in params["states"].items()])

        # Prepare
        engine_states = []
        for _name, obj in params["objects"].items():
            if "engine_states" not in obj:
                continue
            for _cname, s in obj["engine_states"].items():
                engine_states.append(s)
        engine_states = tuple(engine_states)

        return rate, inputs, outputs, states, engine_states, node_names, target_addresses, node

    def _shutdown(self):
        self.backend.logdebug(f"[{self.name}] RxEngine._shutdown() called.")
        self.init_pub.unregister()

    def node_shutdown(self):
        if not self.has_shutdown:
            self.backend.logdebug(f"[{self.name}] RxEngine.node_shutdown() called.")
            for address, node in self.engine.launch_nodes.items():
                self.backend.loginfo(f"[{self.name}] Send termination signal to '{address}'.")
                node.terminate()
            for _, rxnode in self.engine.sp_nodes.items():
                rxnode: RxNode
                if not rxnode.has_shutdown:
                    self.backend.loginfo(f"[{self.name}] Shutting down '{rxnode.name}'.")
                    rxnode.node_shutdown()
            self.backend.loginfo(f"[{self.name}] Shutting down.")
            self._shutdown()
            self.engine.shutdown()
            self.mb.shutdown()
            self.backend.shutdown()
            self.has_shutdown = True


if __name__ == "__main__":
    # Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("-b", "--backend", nargs=1, help="The selected backend (format: '<module>/<BackendClass>').", type=str)
    parser.add_argument("-l", "--loglevel", nargs=1, help="The log level for the environment.", type=int, default=30)
    parser.add_argument("-e", "--env", nargs=1, help="The environment name.", type=str)
    parser.add_argument("-n", "--name", nargs=1, help="The node name.", type=str)
    parser.add_argument(
        "-o",
        "--object",
        nargs=1,
        help="The object name if the node is part of an engine-specific implementation.",
        type=str,
        default=[""],
    )
    args, unknown = parser.parse_known_args()
    backend_id = args.backend[0]
    log_level = args.loglevel[0]
    ns = f"/{args.env[0]}"
    name = args.name[0]
    object_name = args.object[0]

    backend = Backend.from_cmd(ns, backend_id, log_level)

    message_broker = eagerx.core.rx_message_broker.RxMessageBroker(owner=f"{ns}/{name}", backend=backend)

    pnode = None

    try:
        pnode = RxEngine(name=f"{ns}/{name}", message_broker=message_broker)

        message_broker.connect_io()

        pnode.node_initialized()

        backend.spin()
    finally:
        if pnode is not None and not pnode.has_shutdown:
            backend.loginfo(f"Terminating '{ns}/{name}'")
            pnode.node_shutdown()



from eagerx import Graph
from eagerx.core.specs import EngineSpec, BackendSpec
from eagerx.core.env import BaseEnv
from typing import Dict, Tuple
import gymnasium as gym


class EagerxGym(BaseEnv):
    def __init__(
        self,
        name: str,
        rate: float,
        graph: Graph,
        engine: EngineSpec,
        backend: BackendSpec,
        force_start: bool = True,
        render_mode: str = None,
    ):
        super(EagerxGym, self).__init__(name, rate, graph, engine, backend, force_start=force_start, render_mode=render_mode)

    @property
    def observation_space(self) -> gym.spaces.Dict:
        space = self._observation_space
        space.pop("reward")
        space.pop("terminated")
        space.pop("truncated")
        return space

    @property
    def action_space(self) -> gym.spaces.Dict:
        return self._action_space

    def reset(self, seed=None, options=None) -> Tuple[Dict, Dict]:
        # Sample desired states
        states = self.state_space.sample()

        # Perform reset
        obs = self._reset(states)
        obs.pop("reward")
        obs.pop("terminated")
        obs.pop("truncated")

        # Render
        if self.render_mode == "human":
            self.render()
        return obs, {}

    def step(self, action: Dict) -> Tuple[Dict, float, bool, bool, Dict]:
        obs = self._step(action)
        # Render
        if self.render_mode == "human":
            self.render()
        return obs, obs.pop("reward"), bool(obs.pop("terminated")), bool(obs.pop("truncated")), dict()



# import dependencies
try:
    from google.colab.output import eval_js
    from IPython.display import display, Javascript
except ImportError:
    Warning(
        "It seems that you are trying to use this package outside of  google colab. Using a mock version instead, "
        "that does not do anything."
    )

    def mock_fn(*args, **kwargs):
        pass

    eval_js = mock_fn
    display = mock_fn
    Javascript = mock_fn

from base64 import b64encode
import numpy as np
from collections import deque
from typing import List, Union, Optional
import PIL
import io
import threading
import time


class InlineRender:
    def __init__(self, fps: int, maxlen: int = 200, shape: Optional[List[int]] = None, maxwidth: int = 300):
        self.fps = fps
        self.buffer = deque(maxlen=maxlen)
        self.maxlen = maxlen
        self.shape = shape if isinstance(shape, list) else [28, 28]
        self.timestamp = time.time()
        interval = max(1, int(1000 / fps))
        js = Javascript(
            """
        const queue = []
        // Frame
        const div = document.createElement('div');
        div.style.border = '2px solid black';
        div.style.padding = '3px';
        div.style.width = '100%';
        div.style.maxWidth = 'MAXWIDTHpx';
        document.body.appendChild(div);
        // Video
        const img = document.createElement('img');
        img.id = 'my_video'
        img.style.display = 'block';
        img.width = div.clientWidth - 6;
        div.appendChild(img)
        setInterval(() => {
          const n_img = queue.pop()
          if (n_img) {
            img.src = n_img;
          }
        }, 1/FPS) // 1/FPS is replaced with interval argument
        // Synchronize button
        const capture = document.createElement('button');
        capture.textContent = 'Synchronize';
        div.appendChild(capture);
        capture.onclick = function(){
          queue.length= 0
          };
        // Add frames to buffer
        function push_images(...images) {
          // images.forEach(image => queue.push(image))
          images.forEach(image => queue.unshift(image))
        }
    """.replace(
                "1/FPS", str(interval)
            ).replace(
                "MAXWIDTH", str(maxwidth)
            )
        )
        display(js)

        # Create async push loop
        self.event = threading.Event()
        thread = threading.Thread(target=self._async_pushing, args=())
        thread.start()

    def _async_pushing(self):
        """Asynchronously push images from client (server that hosts Python) to local browser (js)."""
        while True:
            self.event.wait()  # Wait for event (ie new buffered images)
            self.event.clear()  # Clear event
            self._push()  # Push buffered images

    def _push(self):
        """Push buffered images from client (server that hosts Python) to local browser (js)."""
        # print("pushing")
        msg = "push_images("
        for _ in range(len(self.buffer)):
            img = self.buffer.popleft()
            msg += '"{}",'.format(img)
        eval_js(msg[:-1] + ")", ignore_result=True, timeout_sec=1.0)
        # print("finished")

    def _img_to_bytes(self, img: np.ndarray):
        """Convert image to JS format

        :params bbox_array: Numpy array (pixels) containing rectangle to overlay on video stream.
        :returns: Base64 image byte string.
        """
        # convert array into PIL image
        img_PIL = PIL.Image.fromarray(img).convert("RGB")
        # resize image if not matching desired self.shape (defined in .yaml)
        if img.shape[:2] != tuple(self.shape):
            img_PIL.thumbnail(self.shape, PIL.Image.ANTIALIAS)
        # format bbox into png for return
        iobuf = io.BytesIO()
        img_PIL.save(iobuf, format="png")
        # format return string
        img_bytes = "data:image/png;base64,{}".format((str(b64encode(iobuf.getvalue()), "utf-8")))
        return img_bytes

    def buffer_images(self, images: Union[List[np.ndarray], np.ndarray]):
        """Buffer images at client side (server that hosts Python)."""
        self.timestamp = time.time()
        if not isinstance(images, list):
            images = [images]
        # Add images to buffer
        for img in images:
            self.buffer.append(self._img_to_bytes(img))
        # Signal push thread that new images are buffered
        self.event.set()
        # Check for buffer overflow
        if len(self.buffer) == self.maxlen:
            return False  # Buffer about to overflow
        else:
            return True  # Buffer not full.



import copy
import numpy as np
import typing
import time
import threading
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count
from eagerx.utils.utils import Header
from eagerx.core.pubsub import Publisher, Subscriber, ShutdownService
import eagerx
from eagerx.core.constants import (
    WARN,
    BackendException,
    Unspecified,
)


# A singleton that is used to check if an argument was specified.
_unspecified = Unspecified()


def merge(a: typing.Dict, b: typing.Dict, path=None):
    """merges b into a"""
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


def split(a: typing.Any):
    if isinstance(a, dict):
        for key in list(a):
            value = a.pop(key)
            value = split(value)
            keys = [k for k in key.split("/") if len(k) > 0]
            for kk in reversed(keys[1:]):
                value = {kk: value}
            if keys[0] not in a:
                a[keys[0]] = value
            else:
                merge(a[keys[0]], value)
    return a


class SingleProcess(eagerx.Backend):

    BACKEND = "SINGLE_PROCESS"
    DISTRIBUTED_SUPPORT = False
    MULTIPROCESSING_SUPPORT = False
    COLAB_SUPPORT = True

    MIN_THREADS = 10

    @classmethod
    def make(cls, log_level=WARN) -> eagerx.specs.BackendSpec:
        spec = cls.get_specification()
        spec.config.log_level = log_level if isinstance(log_level, str) else eagerx.get_log_level()
        return spec

    def initialize(self):
        self._backend = self
        self._pserver = dict()
        self._topics = dict()
        self._cond = threading.Condition()
        self._tpool = ThreadPoolExecutor(max_workers=max(self.MIN_THREADS, cpu_count()))

    def Publisher(self, address: str, dtype: str):
        return _Publisher(self._backend, self._tpool, self._topics, self._cond, address, dtype)

    def Subscriber(self, address: str, dtype: str, callback, header: bool = False, callback_args=tuple()):
        return _Subscriber(
            self._backend, self._topics, self._cond, address, dtype, callback, header, callback_args=callback_args
        )

    def register_environment(self, name: str, force_start: bool, fn: typing.Callable):
        return _ShutdownService()

    def delete_param(self, param: str, level: int = 1) -> None:
        try:
            keys = [k for k in param.split("/") if len(k) > 0]
            val = self._pserver
            for key in keys[:-1]:
                val = val[key]
            val.pop(keys[-1])
            self.loginfo(f'Parameters under namespace "{param}" deleted.')
        except KeyError as e:
            if level == 0:
                raise BackendException(e)
            elif level == 1:
                self.logwarn(e)
            else:
                pass

    def upload_params(self, ns: str, values: typing.Dict, verbose: bool = False) -> None:
        values = copy.deepcopy(values)
        ns = [k for k in ns.split("/") if len(k) > 0]
        ns_values = split(values)
        for k in ns:
            ns_values = {k: ns_values}
        merge(self._pserver, ns_values)

    def get_param(self, name: str, default: typing.Any = _unspecified):
        try:
            keys = [k for k in name.split("/") if len(k) > 0]
            val = self._pserver
            for key in keys:
                val = val[key]
            return val
        except KeyError as e:
            if not isinstance(default, Unspecified):
                return default
            else:
                raise BackendException(e)

    def spin(self):
        raise NotImplementedError(f"Not implemented, because backend '{self.BACKEND}' does not support multiprocessing.")

    def shutdown(self) -> None:
        if not self._has_shutdown:
            self.logdebug("Backend.shutdown() called.")
            self._has_shutdown = True
            self._tpool.shutdown(wait=True)


class _Publisher(Publisher):
    def __init__(
        self,
        backend: eagerx.Backend,
        tpool: ThreadPoolExecutor,
        topics: typing.Dict,
        cond: threading.Condition,
        address: str,
        dtype: str,
    ):
        super().__init__(backend)
        self._tpool = tpool
        self._cond = cond
        self._topics = topics
        with self._cond:
            if address not in topics:
                self._topic = dict(pubs=0, subs=[], latched=None, dtype=dtype)
                topics[address] = self._topic
            else:
                self._topic = topics[address]
                assert self._topic["dtype"] == dtype, f"Dtypes do not match for topic {address}."

            # Increase publisher count
            self._topic["pubs"] += 1

            self._address = address
            self._dtype = dtype
            self._name = f"{self._address}"
            self._unregistered = False

    def _publish(self, msg: typing.Union[float, bool, int, str, np.ndarray, np.number], header: Header) -> None:
        if not self._unregistered:
            # todo: check if dtype(msg) == self._dtype?
            # Convert python native types to numpy arrays.
            if isinstance(msg, float):
                msg = np.array(msg, dtype="float32")
            elif isinstance(msg, int) and not isinstance(msg, bool):
                msg = np.array(msg, dtype="int64")

            # Check if message complies with space
            if not isinstance(msg, (np.ndarray, np.number, str, bool)):
                self._backend.logerr(f"[publisher][{self._name}]: type(recv)={type(msg)}")
                time.sleep(10000000)

            # with self._cond:  # todo: needed?
            for cb in self._topic["subs"]:
                self._tpool.submit(cb, msg, header)
            self._topic["latched"] = msg, header

    def unregister(self) -> None:
        if not self._unregistered:
            with self._cond:
                self._unregistered = True
                assert self._topic["pubs"] > 0, "According to the counter, there should be no publishers left for this topic."
                self._topic["pubs"] -= 1

                # If no other subscribers or publishers, remove topic.
                if len(self._topic["subs"]) == 0 and self._topic["pubs"] == 0:
                    self._topics.pop(self._address)


class _Subscriber(Subscriber):
    def __init__(
        self,
        backend: eagerx.Backend,
        topics: typing.Dict,
        cond: threading.Condition,
        address: str,
        dtype: str,
        callback,
        header: bool,
        callback_args=tuple(),
    ):
        super().__init__(backend, header)
        self._cond = cond
        self._topics = topics
        with self._cond:
            if address not in topics:
                self._topic = dict(pubs=0, subs=[], latched=None, dtype=dtype)
                topics[address] = self._topic
                latched = None
            else:
                self._topic = topics[address]
                assert self._topic["dtype"] == dtype, f"Dtypes do not match for topic {address}."
                latched = self._topic["latched"]

            self._unregistered = False
            self._topic["subs"].append(self.callback)
            self._address = address
            self._dtype = dtype
            self._cb_wrapped = callback
            self._cb_args = callback_args
            self._name = f"{self._address}"

        if latched is not None:
            self._backend.logdebug(f"LATCHED: {self._address}")
            self.callback(*latched)  # todo: inside cond?

    def callback(self, msg, header):
        # todo: pass header to callback in publisher
        # todo: pass header to wrapped callback if self._header
        if not self._unregistered:
            # todo: check if dtype(msg) == self._dtype?
            if not isinstance(msg, (np.ndarray, np.number, str, bool)):
                self._backend.logerr(f"[subscriber][{self._name}]: type(recv)={type(msg)}")
                time.sleep(10000000)
            self._cb_wrapped(msg, header, *self._cb_args) if self._header else self._cb_wrapped(msg, *self._cb_args)

    def unregister(self) -> None:
        if not self._unregistered:
            with self._cond:
                self._unregistered = True

                self._topic["subs"] = [cb for cb in self._topic["subs"] if not id(cb) == id(self.callback)]

                # If no other subscribers or publishers, remove topic.
                if len(self._topic["subs"]) == 0 and self._topic["pubs"] == 0:
                    self._topics.pop(self._address)


class _ShutdownService(ShutdownService):
    def __init__(self):
        pass

    def unregister(self):
        pass



import inspect
import functools
from textwrap import indent
from eagerx.core import register
from eagerx.core.entities import Object


def get_info(cls, methods=None, no_cls=False, return_msg=False):
    """Get information on the entity's registered function"""
    REGISTRY = register.REGISTRY
    TYPE_REGISTER = register.TYPE_REGISTER

    id = cls.__module__ + "/" + cls.__qualname__

    # Determine methods to print info on
    if methods is None:
        methods = list(cls.INFO.keys())
    elif isinstance(methods, str):
        methods = [methods]
    assert isinstance(
        methods, list
    ), "Incorrect specification. Please provide the method as type(methods) =`List[str]`or `str`."

    # Create base info message
    tab = "   "
    msg = indent(f"entity_type: `{cls.__qualname__}`\n", tab)
    msg += indent(f"module: `{cls.__module__}`\n", tab)
    try:
        msg += indent(f"file: `{inspect.getfile(cls)}`\n", tab)
    except TypeError:
        msg += indent("file: <unknown>\n", tab)
    msg += "\n"

    # Objects: Add supported engines
    if issubclass(cls, Object):
        if id not in REGISTRY or REGISTRY[id].keys() == 0:
            engine_msg = "Supported engines: <Nothing registered>\n"
        else:
            engine_msg = "Supported engines:\n"
            for engine_id in REGISTRY[id].keys():
                if engine_id in ["make", "cls"]:
                    continue
                engine_msg += indent(f"{engine_id}\n", tab[2:] + "- ")

        msg += engine_msg + "\n"

    # Make spec message
    method_fn = cls.make
    sig = inspect.signature(method_fn)
    msg += "Make this spec with:\n"
    make_msg = f"spec = {method_fn.__qualname__}{sig}\n"
    msg += indent(make_msg + "\n", tab)

    # Generate class
    if not no_cls:
        msg += f"class {cls.__qualname__}:\n"
        msg += indent(cls.__doc__ + "\n\n", tab) if isinstance(cls.__doc__, str) else ""

        # Expand info message
        for method, register_fns in cls.INFO.items():
            if method not in methods:
                continue
            method_fn = getattr(cls, method)
            sig = inspect.signature(method_fn)
            method_msg = f"{method}{sig}:\n"

            # Replace if wrapped
            if hasattr(method_fn, "__wrapped__") and isinstance(method_fn.__wrapped__, functools.partial):
                method_fn = method_fn.__wrapped__.func

            for r in register_fns:
                try:
                    args = TYPE_REGISTER[id][r]
                    if len(args) == 0:
                        registered_msg = f"{r}: <Nothing registered>\n"
                    else:
                        registered_msg = f"{r}:\n"
                    for name, value in args.items():
                        registered_msg += indent(f"{name}: {value}\n", tab[2:] + "- ")
                except KeyError:
                    registered_msg = f"{r}: <Nothing registered>\n"
                method_msg += indent(registered_msg, tab)

            if isinstance(method_fn.__doc__, str):
                docs_msg = "docs:\n"
                docs_msg += indent(method_fn.__doc__ + "\n", tab)
            else:
                docs_msg = "docs: <Not available>\n"
            method_msg += indent(docs_msg, tab)

            msg += indent(method_msg + "\n", tab)

    # Print message
    if return_msg:
        return msg
    else:
        print(msg)
